//+-----------------------------------------------------------------------------+
//| Copyright (C) 2013 REGEO (http://www.regeo.uji.es/)
//|
//| This program is free software
//| You can redistribute it and/or modify it under the terms of the GNU General Public License,
//| as published by the Free Software Foundation.
//| 
//| This program is distributed in the hope that it will be useful,
//| but WITHOUT ANY WARRANTY; without even the implied warranty of
//| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//+----------------------------------------------------------------------------*/

#include <limits>   //_isnan, numeric_limits, max_exponent
#include <cmath>    // sqrt
#include <algorithm>
#include <vector>
#include <complex>
#include "Tools_Geometry.h"
#include "Tools_Matrix.h"

#include "Arc.h"
#include "Stroke.h"
#include "FitEllipse.h"

CFitEllipse::CFitEllipse()
{
	ArcTolMin= 0.010;
	ArcTolMax= 0.100;
	ArcSmoothingPenalty= 0.05;

	StrokeCornersApproach = 0;   // 0= automatic, 1= IStraw, 2= Silding Strips;
	FitEllipseApproach= 0; // 0= automatic, 1= 5P, 2= DIR, 3= GEF
	FitEllipseSpeed= 0; // 0= fast, 1= balance, 2= accurate
	UseConvexHull5P= true;
	CheckQuality5P= true;
	CheckQualityDIR= true;
	ResampleStrokes= true;
	NumResampledPoints= 20;
	ResampleStrokesGEF= true;
	NumResampledPointsGEF= 20;
	DisplayColours= false;
}


CFitEllipse::CFitEllipse(double ATMin, double ATMax, double SPenalty,
						 long SCApproach,
						 long FEApproach,
						 long FESpeed,
						 bool UseCH5P,
						 bool Quality5P, bool QualityDIR,
						 bool RStrokes, long NumP,
						 bool RStrokesGEF, long NumPGEF, bool DColours)
//Initialize with input parameters
: ArcTolMin(ATMin),
  ArcTolMax (ATMax),
  ArcSmoothingPenalty (SPenalty),
  StrokeCornersApproach (SCApproach),
  FitEllipseApproach (FEApproach),
  FitEllipseSpeed (FESpeed),
  UseConvexHull5P (UseCH5P),
  CheckQuality5P (Quality5P),
  CheckQualityDIR (QualityDIR),
  ResampleStrokes (RStrokes),
  NumResampledPoints (NumP),
  ResampleStrokesGEF (RStrokesGEF),
  NumResampledPointsGEF (NumPGEF),
  DisplayColours (DColours)
{
}


CFitEllipse::~CFitEllipse()
{
}


bool CFitEllipse::IsEllipse(CConicParams phi)
//The input is the set of parameters of a general quadratic curve
//phi (a, b, c, d, e, f),
//where ax^2+bxy+cy^2+dx+ey+f=0

//Returns true if the parameters depict and ellipse, false otherwise
{
	//Determinant |a b/2 d/2; b/2 c e/2; d/2 e/2 f|
	double Delta = phi.a*phi.c*phi.f - phi.b*phi.b*phi.f/4 +phi.b*phi.d*phi.e/4 - phi.c*phi.d*phi.d/4 - phi.a*phi.e*phi.e/4;
	double J = phi.a*phi.c - phi.b*phi.b/4; //|a b/2; b/2 c|

	//Degenerate conic iff Delta==0
	//Hyperbola iff J<0
	//Parabola iff J==0
	//Ellipse iff J>0
	//Imaginary ellipse iff Delta*(a+c)>0

	return(Delta!=0 && J>0 && Delta*(phi.a+phi.c)<0);
}


bool CFitEllipse::FivePointsConic(double p0[3], double p1[3], double p2[3],
								  double p3[3], double p4[3], 
								  CConicParams &phi)

// Code to find the equation of a conic */
//               Tom Davis              */
//             April 12, 1996           */
// http://mathforum.org/library/drmath/view/51735.html

// takes five points in homogeneous coordinates, and returns the
// coefficients of a general conic equation in a, b, c, ..., f:
//
// a*x*x + b*x*y + c*y*y + d*x + e*y + f = 0.
//
// The routine returns 1 on success; 0 otherwise.  (It can fail, for
// example, if there are duplicate points.
//
// Typically, the points will be finite, in which case the third (w)
// coordinate for all the input vectors will be 1, although the code
// deals cleanly with points at infinity.
//
// For example, to find the equation of the conic passing through
// (5, 0), (-5, 0), (3, 2), (3, -2), and (-3, 2), set:
//
// p0[0] =  5, p0[1] =  0, p0[2] = 1,
// p1[0] = -5, p1[1] =  0, p1[2] = 1,
// p2[0] =  3, p2[1] =  2, p2[2] = 1,
// p3[0] =  3, p3[1] = -2, p3[2] = 1,
// p4[0] = -3, p4[1] =  2, p4[2] = 1.
//
// But if you want the equation of the hyperbola that is tangent to the
// line 2x=y at infinity, simply make one of the points be the point at
// infinity along that line, for example:
//
// p0[0] = 1, p0[1] = 2, p0[2] = 0.
{
	double L0[3], L1[3], L2[3], L3[3], Q[3];
	cross(p0, p1, L0)
	cross(p1, p2, L1)
	cross(p2, p3, L2)
	cross(p3, p4, L3)
	cross(L0, L3, Q)

	double A = Q[0];double B = Q[1];double C = Q[2];
	double a1 = L1[0];double b1 = L1[1];double c1 = L1[2];
	double a2 = L2[0];double b2 = L2[1];double c2 = L2[2];
	double x0 = p0[0];double y0 = p0[1];double w0 = p0[2];
	double x4 = p4[0];double y4 = p4[1];double w4 = p4[2];

	double aa = -A*a1*a2*y4*w0
				+A*a1*a2*w4*y0
				-B*b1*a2*y4*w0
				-B*c1*a2*w4*w0
				+B*a1*b2*w4*y0
				+B*a1*c2*w4*w0
				+C*b1*a2*y4*y0
				+C*c1*a2*w4*y0
				-C*a1*b2*y4*y0
				-C*a1*c2*y4*w0;

	double cc =  A*c1*b2*w4*w0
				+A*a1*b2*x4*w0
				-A*b1*c2*w4*w0
				-A*b1*a2*w4*x0
				+B*b1*b2*x4*w0
				-B*b1*b2*w4*x0
				+C*b1*c2*x4*w0
				+C*b1*a2*x4*x0
				-C*c1*b2*w4*x0
				-C*a1*b2*x4*x0;

	double ff =  A*c1*a2*y4*x0
				+A*c1*b2*y4*y0
				-A*a1*c2*x4*y0
				-A*b1*c2*y4*y0
				-B*c1*a2*x4*x0
				-B*c1*b2*x4*y0
				+B*a1*c2*x4*x0
				+B*b1*c2*y4*x0
				-C*c1*c2*x4*y0
				+C*c1*c2*y4*x0;

	double bb =  A*c1*a2*w4*w0
				+A*a1*a2*x4*w0
				-A*a1*b2*y4*w0
				-A*a1*c2*w4*w0
				-A*a1*a2*w4*x0
				+A*b1*a2*w4*y0
				+B*b1*a2*x4*w0
				-B*b1*b2*y4*w0
				-B*c1*b2*w4*w0
				-B*a1*b2*w4*x0
				+B*b1*b2*w4*y0
				+B*b1*c2*w4*w0
				-C*b1*c2*y4*w0
				-C*b1*a2*x4*y0
				-C*b1*a2*y4*x0
				-C*c1*a2*w4*x0
				+C*c1*b2*w4*y0
				+C*a1*b2*x4*y0
				+C*a1*b2*y4*x0
				+C*a1*c2*x4*w0;

	double dd = -A*c1*a2*y4*w0
				+A*a1*a2*y4*x0
				+A*a1*b2*y4*y0
				+A*a1*c2*w4*y0
				-A*a1*a2*x4*y0
				-A*b1*a2*y4*y0
				+B*b1*a2*y4*x0
				+B*c1*a2*w4*x0
				+B*c1*a2*x4*w0
				+B*c1*b2*w4*y0
				-B*a1*b2*x4*y0
				-B*a1*c2*w4*x0
				-B*a1*c2*x4*w0
				-B*b1*c2*y4*w0
				+C*b1*c2*y4*y0
				+C*c1*c2*w4*y0
				-C*c1*a2*x4*y0
				-C*c1*b2*y4*y0
				-C*c1*c2*y4*w0
				+C*a1*c2*y4*x0;

	double ee = -A*c1*a2*w4*x0
				-A*c1*b2*y4*w0
				-A*c1*b2*w4*y0
				-A*a1*b2*x4*y0
				+A*a1*c2*x4*w0
				+A*b1*c2*y4*w0
				+A*b1*c2*w4*y0
				+A*b1*a2*y4*x0
				-B*b1*a2*x4*x0
				-B*b1*b2*x4*y0
				+B*c1*b2*x4*w0
				+B*a1*b2*x4*x0
				+B*b1*b2*y4*x0
				-B*b1*c2*w4*x0
				-C*b1*c2*x4*y0
				+C*c1*c2*x4*w0
				+C*c1*a2*x4*x0
				+C*c1*b2*y4*x0
				-C*c1*c2*w4*x0
				-C*a1*c2*x4*x0;

	if (aa != 0.0) {
		bb /= aa; cc /= aa; dd /= aa; ee /= aa; ff /= aa; aa = 1.0;
	}
	else if (bb != 0.0) {
		cc /= bb; dd /= bb; ee /= bb; ff /= bb; bb = 1.0;
	}
	else if (cc != 0.0) {
		dd /= cc; ee /= cc; ff /= cc; cc = 1.0;
	}
	else if (dd != 0.0) {
		ee /= dd; ff /= dd; dd = 1.0;
	}
	else if (ee != 0.0) {
		ff /= ee; ee = 1.0;
	}
	else {
		return false;
	}

	phi.PutParams(aa, bb, cc, dd, ee, ff);

	return IsEllipse(phi);
}


bool CFitEllipse::FitEllipse5P(CStroke InputStroke, CConicParams &phi)
//Returns the ellipse that fits the two endpoints plus three intermediate points

//Use the first endpoint plus four intermediate points
//if both endpoints are too close
{
	long n= InputStroke.GetNumPoints()-1;

	//Select five equally spaced points
	//---------------------------------
	POINT2D P0= InputStroke.GetPoint(0).GetPosition();
	POINT2D P4= InputStroke.GetPoint(n).GetPosition();

	if(MODULE_2D(P0, P4) < 0.1*InputStroke.GetLength())
		//Use four intermediate points if endpoints are too close to each other
		InputStroke.Resample(6);
	else
		InputStroke.Resample(5);
	//Select the first, the last and three equally spaced points
	//P0= InputStroke.GetPoint(0).GetPosition();
	POINT2D P1= InputStroke.GetPoint(1).GetPosition();
	POINT2D P2= InputStroke.GetPoint(2).GetPosition();
	POINT2D P3= InputStroke.GetPoint(3).GetPosition();
	P4= InputStroke.GetPoint(4).GetPosition();

	//Fit the ellipse passing through five points
	//-------------------------------------------
	double p0[3], p1[3], p2[3], p3[3], p4[3];
	p0[0] = P0.x; p0[1] = P0.y; p0[2] = 1;
	p1[0] = P1.x; p1[1] = P1.y; p1[2] = 1;
	p2[0] = P2.x; p2[1] = P2.y; p2[2] = 1;
	p3[0] = P3.x; p3[1] = P3.y; p3[2] = 1;
	p4[0] = P4.x; p4[1] = P4.y; p4[2] = 1;

	return FivePointsConic(p0, p1, p2, p3, p4, phi);
}


bool CFitEllipse::FitEllipse5PDQ(CStroke InputStroke, CConicParams &phi)
//Returns the ellipse that fits the two endpoints plus three intermediate points

//	WARNING! Not valid for closed strokes (where both endpoints are too close)
{
	long n= InputStroke.GetNumPoints()-1;

	//Select five Drawn and Quartered points
	//--------------------------------------
	//Take points P0 and P4 as the two end-points of the line
	POINT2D P0= InputStroke.GetPoint(0).GetPosition();
	POINT2D P4= InputStroke.GetPoint(n).GetPosition();

	if(MODULE_2D(P0, P4) < 0.1*InputStroke.GetLength())
		//Approach not valid for closed strokes
		return false;

	//Take point P2 as the point furthest from the line P0P4
	POINT2D P2= InputStroke.GetPoint(1).GetPosition();
	double mindist= GetDistancePointToSegment(P0, P4, P2);
	long c= 1;
	for (long i=1; i<n-1; i++){
		POINT2D pp= InputStroke.GetPoint(i).GetPosition();
		double dist= GetDistancePointToSegment(P0, P4, pp);
		if(dist > mindist){
			mindist= dist;
			c= i;
			P2= pp;
		}
	}

	//Take point P1 as the point between P0 and P2 furthest from the line PoP2
	POINT2D P1= InputStroke.GetPoint(1).GetPosition();
	mindist= GetDistancePointToSegment(P0, P2, P1);
	long d= 1;
	for (long i=1; i<c-1; i++){
		POINT2D pp= InputStroke.GetPoint(i).GetPosition();
		double dist= GetDistancePointToSegment(P0, P2, pp);
		if(dist > mindist){
			mindist= dist;
			d= i;
			P1= pp;
		}
	}

	//Take point E as the point between C and B furthest from the line CB
	POINT2D P3= InputStroke.GetPoint(c+1).GetPosition();
	mindist= GetDistancePointToSegment(P2, P4, P3);
	//long e= 1;
	for (long i=c+2; i<n-1; i++){
		POINT2D pp= InputStroke.GetPoint(i).GetPosition();
		double dist= GetDistancePointToSegment(P2, P4, pp);
		if(dist > mindist){
			mindist= dist;
			//e= i;
			P3= pp;
		}
	}

	//Fit the ellipse passing through five points
	//-------------------------------------------
	double p0[3], p1[3], p2[3], p3[3], p4[3];
	p0[0] = P0.x; p0[1] = P0.y; p0[2] = 1;
	p1[0] = P1.x; p1[1] = P1.y; p1[2] = 1;
	p2[0] = P2.x; p2[1] = P2.y; p2[2] = 1;
	p3[0] = P3.x; p3[1] = P3.y; p3[2] = 1;
	p4[0] = P4.x; p4[1] = P4.y; p4[2] = 1;

	return FivePointsConic(p0, p1, p2, p3, p4, phi);
}


bool CFitEllipse::FitEllipseDIR(CStroke InputStroke, CConicParams &phi)
//Returns the coefficients of the ellipse that best fits the points of InputStroke

//Returns true if the stroke fits an ellipse
//The output parameters belong to the general quadratic curve ax^2+bxy+cy^2+dx+ey+f=0

//Note that all data points are assumed to belong to one single ellipse

// The function follows the approach by Fitzgibbon, as implemented by:
//Radim Halir , Jan Flusser. (1998). Numerically Stable Direct Least Squares Fitting Of Ellipses. WSCG98

// A C# implementation can be found at:
//http://skotagiri.wordpress.com/2010/06/19/c-implementation-for-fitting-an-ellipse-for-a-set-of-points/

// Warning: This fit is somewhat biased toward smaller ellipses
{
	int numPoints = (long)InputStroke.GetNumPoints();

	//C1= First quarter of constraint matrix
	std::vector<std::vector<double>> C1(3, std::vector<double>(3, 0));
	C1[0][0] = 0;     C1[0][1] = 0;     C1[0][2] = 0.5;
	C1[1][0] = 0;     C1[1][1] = -1;    C1[1][2] = 0;
	C1[2][0] = 0.5;   C1[2][1] = 0;     C1[2][2] = 0;

	//2 D1 = quadratic part of the design matrix
	//3 D2 = linear part of the design matrix
	std::vector<std::vector<double>> D1(numPoints, std::vector<double>(3, 0));
	std::vector<std::vector<double>> D2(numPoints, std::vector<double>(3, 0));

	for (long xx=0; xx<numPoints; xx++){
		POINT2D pp= InputStroke.GetPoint(xx).GetPosition();

		D1[xx][0] = pp.x * pp.x;
		D1[xx][1] = pp.x * pp.y;
		D1[xx][2] = pp.y * pp.y;

		D2[xx][0] = pp.x;
		D2[xx][1] = pp.y;
		D2[xx][2] = 1;
	}

	//Scatter matrix
	//Quadratic part of the scatter matrix
	std::vector<std::vector<double>> S1(3, std::vector<double>(3, 0));
	S1= MProduct(MTranspose(D1), D1);

	//Combined part of the scatter matrix
	std::vector<std::vector<double>> S2(3, std::vector<double>(3, 0));
	S2= MProduct(MTranspose(D1), D2);

	//Linear part of the scatter matrix
	std::vector<std::vector<double>> S3(3, std::vector<double>(3, 0));
	S3= MProduct(MTranspose(D2), D2);

	//T = for getting a2 from a1
	std::vector<std::vector<double>> T(3, std::vector<double>(3, 0));
	T= MScale(MProduct(MInverse(S3), MTranspose(S2)), -1);

	//M = reduced scatter matrix
	std::vector<std::vector<double>> M(3, std::vector<double>(3, 0));
	M= MAdd(S1, MProduct(S2, T));

	//9 M = [M(3, :) ./ 2; - M(2, :); M(1, :) ./ 2]; % premultiply by inv(C1)
	////M = C1 * M;
	M= MProduct(MInverse(C1), M);

	//Find the appropriate eigenvector a1 of the matrix M
	std::vector<std::complex<double>> eigenvalues(3, (1.0,0.0));
	std::vector<std::vector<double>> EigenVectors(3, std::vector<double>(3, 0));
	Eigen eigen;
	if(!eigen.UnsymEigen(M, eigenvalues, EigenVectors))
		return false;   //Unable to find eigenvalues (probably the stroke is a perfect straigth line)

	//Save eigenvectors in complex format
	std::vector<std::vector<std::complex<double>>> cev(3, std::vector<std::complex<double>>(3, (0.0,0.0)));

	for (long i= 0; i<3; i++){
		if(std::imag(eigenvalues[i]) != 0){
			//Complex eigenvalues of a real matrix are in conjugate pairs
			for (long j= 0; j<3; j++)
				cev[i][j]= std::complex<double>(EigenVectors[j][i], EigenVectors[j][i+1]);

			for (long j= 0; j<3; j++)
				cev[i][j]/= cev[i][2];

			for (long j= 0; j<3; j++)
				cev[i+1][j]= std::conj(cev[i][j]);
			i++;
		}
		else{
			for (long j= 0; j<3; j++)
				cev[i][j]= std::complex<double>(EigenVectors[j][i], 0);

			for (long j= 0; j<3; j++)
				cev[i][j]/= cev[i][2];
		}
	}

	//11 cond = 4 * evec(1, :) .* evec(3, :) - evec(2, :) . 2; % evaluate aCa
	//12 a1 = evec(:, find(cond > 0)); % eigenvector for min. pos. eigenvalue
	std::vector<std::vector<double>> a1(3, std::vector<double>(1, 0));
	for (int xx = 0; xx < 3; xx++){
		std::complex<double> condition = std::complex<double> (4,0) * cev[xx][0] * cev[xx][2] - cev[xx][1] * cev[xx][1];
		if (std::imag(condition) == 0 && std::real(condition) > 0){
			// Solution is found
			for (int yy = 0; yy < 3; yy++){
				a1[yy][0] = std::real(cev[xx][yy]);
			}
		}
	}

	//13 a2 = T * a1; % ellipse coefficients
	std::vector<std::vector<double>> a2(3, std::vector<double>(1, 0));
	a2= MProduct(T, a1);

	//Ellipse coefficients of general form
	phi.PutParams(a1[0][0], a1[1][0], a1[2][0], a2[0][0], a2[1][0], a2[2][0]);

	return IsEllipse(phi);
}


void CFitEllipse::normalise2dpts(std::vector<std::vector<double>> &pts,
								 std::vector<std::vector<double>> &T)
// Translates and normalises a set of 2D homogeneous points 
// so that their centroid is at the origin and their mean distance from 
// the origin is sqrt(2)
// This process typically improves the conditioning of any equations
//  used to solve homographies, fundamental matrices etc.

// Argument:
//   pts -  3xN array of 2D homogeneous coordinates

// Returns:
//   T   -  The 3x3 transformation matrix, newpts = T*pts

//Based on
// Peter Kovesi
// School of Computer Science & Software Engineering
// The University of Western Australia
// http://www.csse.uwa.edu.au/~pk
{
	// Centroid
	double c0= 0.;
	double c1= 0.;
	long NumPts= pts[0].size();
	for(long i=0; i<NumPts; i++){
		c0+= pts[0][i];
		c1+= pts[1][i];
	}

	c0 /= (double)NumPts;
	c1 /= (double)NumPts;

	// Scale so that the mean distance from the centroid to a point equals sqrt(2)
	double meandist= 0.;
	for(long i=0; i<NumPts; i++){
		double pts0i= pts[0][i] - c0;
		double pts1i= pts[1][i] - c1;
		meandist+= sqrt(pts0i*pts0i + pts1i*pts1i);
	}
	meandist/= (double)NumPts;
	double scale = sqrt(2.)/meandist;

	T[0][0] = scale;   T[0][1] = 0.;      T[0][2] = -scale*c0;
	T[1][0] = 0.;      T[1][1] = scale;   T[1][2] = -scale*c1;
	T[2][0] = 0.;      T[2][1] = 0.;      T[2][2] = 1.;
}


void CFitEllipse::LevenbergMarquardtStep(SzpakData &S)
//   Function: levenbergMarquardtStep
//
//   This function is used in the main loop of guaranteedEllipseFit in the process
//   of minimizing an approximate maximum likelihood cost function of an
//   ellipse fit to data.  It computes an update for the parameters
//   representing the ellipse, using the method of Levenberg-Marquardt for
//   non-linear optimisation. 
//   See: http://en.wikipedia.org/wiki/Levenberg//E2//80//93Marquardt_algorithm
//
//   Parameters:
//      S a data structure containing various parameters
//        needed for the optimisation process
//
//   Returns: 
//     the same data structure S, except that relevant fields have been updated
//
//  Based on a matlab code by Zygmunt L. Szpak (c) 2012
{
	// extract variables from data structure
	std::vector<std::vector<double>> t(6, std::vector<double>(1, 0));
	for(long i= 0; i<6; i++)
		t[i][0]= S.t[i][S.k];
	std::vector<std::vector<double>> delta(6, std::vector<double>(1, 0));
	for(long i= 0; i<6; i++)
		delta[i][0]= S.delta[i][S.k];
	double current_cost = S.cost[0][S.k];

	// compute two potential updates for theta based on different weightings of
	// the identity matrix.
	//-------------------------------------------------------------------------

	std::vector<std::vector<double>> jacob;
	jacob= MProduct(MTranspose(MConcatenateRows(S.jacobian_matrix, S.jacobian_matrix_barrier)) , S.r);
	std::vector<std::vector<double>> MtFt;
	MtFt= MProduct(MTranspose(t), MProduct(S.F, t));
	double tFt= MtFt[0][0];

	// We solve for the new update direction in a numerically careful manner
	// If we didn't worry about numerical stability then we would compute 
	// the first new search direction like this:

	std::vector<std::vector<double>> update_a;
	// update_a = MScale(MLeftDivide(MAdd(H, MScale(S.I, S.lambda)), jacob), -1);

	// But close to the barrier between ellipses and hyperbolas we may
	// experience numerical conditioning problems due to the nature of the
	// barrier term itself.  Hence we perform the calculation in a
	// numerically more stable way with 

	std::vector<std::vector<double>> Z00;
	Z00= MAdd(MProduct(MTranspose(S.jacobian_matrix), S.jacobian_matrix), MScale(S.I, S.lambda));
	std::vector<std::vector<double>> Z01;
	Z01= MScale(MProduct(MTranspose(S.jacobian_matrix_barrier), S.jacobian_matrix_barrier), pow(tFt, 4));
	std::vector<std::vector<double>> Z10;
	Z10= S.I;
	std::vector<std::vector<double>> Z11;
	Z11= MScale(S.I, -pow(tFt, 4));
	std::vector<std::vector<double>> Z_a;
	Z_a= MConcatenateRows(MConcatenateColumns(Z00, Z01), MConcatenateColumns(Z10, Z11));

	std::vector<std::vector<double>> zz_a;
	std::vector<std::vector<double>> zeros(6, std::vector<double>(1, 0));
	zz_a= MScale(MConcatenateRows(jacob, zeros), -1);

	update_a = MLeftDivide(Z_a, zz_a);
	// drop the nuisance parameter components
	update_a.resize(6);

	// In a similar fashion, the second potential search direction could be
	// computed like this:

	std::vector<std::vector<double>> update_b;
	// update_b = MScale(MLeftDivide(MAdd(H, MScale(S.I, lambda/v)), jacob), -1);

	// but instead we computed it with
	//std::vector<std::vector<double>> Z00;
	Z00= MAdd(MProduct(MTranspose(S.jacobian_matrix), S.jacobian_matrix), MScale(S.I, S.lambda/S.damping_multiplier));
	//std::vector<std::vector<double>> Z01;
	Z01= MScale(MProduct(MTranspose(S.jacobian_matrix_barrier), S.jacobian_matrix_barrier), pow(tFt, 4));
	//std::vector<std::vector<double>> Z10;
	Z10= S.I;
	//std::vector<std::vector<double>> Z11;
	Z11= MScale(S.I, -pow(tFt, 4));
	std::vector<std::vector<double>> Z_b;
	Z_b= MConcatenateRows(MConcatenateColumns(Z00, Z01), MConcatenateColumns(Z10, Z11));

	std::vector<std::vector<double>> zz_b;
	//std::vector<std::vector<double>> zeros(6, std::vector<double>(1, 0));
	zz_b= MScale(MConcatenateRows(jacob, zeros), -1);

	update_b = MLeftDivide(Z_b, zz_b);
	// drop the nuisance parameter components
	update_b.resize(6);

	// the potential new parameters are then
	std::vector<std::vector<double>> t_potential_a(6, std::vector<double>(1, 0));
	t_potential_a = MAdd(t, update_a);
	std::vector<std::vector<double>> t_potential_b(6, std::vector<double>(1, 0));
	t_potential_b = MAdd(t, update_b);

	// compute new residuals and costs based on these updates
	//-------------------------------------------------------

	// residuals computed on data points
	double cost_a = 0;
	double cost_b = 0;
	for(long i = 0; i<S.numberOfPoints; i++){
		// transformed data point
		std::vector<std::vector<double>> ux_j(6, std::vector<double>(1, 0));
		ux_j[0][0]= S.data_points[0][i]*S.data_points[0][i];
		ux_j[1][0]= S.data_points[0][i]*S.data_points[1][i];
		ux_j[2][0]= S.data_points[1][i]*S.data_points[1][i];
		ux_j[3][0]= S.data_points[0][i];
		ux_j[4][0]= S.data_points[1][i];
		ux_j[5][0]= 1.;

		// derivative of transformed data point
		std::vector<std::vector<double>> dux_j(6, std::vector<double>(2, 0));
		dux_j[0][0]= 2*S.data_points[0][i];
		dux_j[1][0]= S.data_points[1][i];
		dux_j[2][0]= 0.;
		dux_j[3][0]= 1.;
		dux_j[4][0]= 0.;
		dux_j[5][0]= 0.;
		dux_j[0][1]= 0;
		dux_j[1][1]= S.data_points[0][i];
		dux_j[2][1]= 2*S.data_points[1][i];
		dux_j[3][1]= 0.;
		dux_j[4][1]= 1.;
		dux_j[5][1]= 0.;

		// outer product
		std::vector<std::vector<double>> A;
		A = MProduct(ux_j, MTranspose(ux_j));

		// use identity covs
		std::vector<std::vector<double>> B;
		B = MProduct(dux_j, MTranspose(dux_j));

		std::vector<std::vector<double>> Mt_aBt_a;
		Mt_aBt_a= MProduct(MTranspose(t_potential_a), MProduct(B, t_potential_a));
		double t_aBt_a= Mt_aBt_a[0][0];
		std::vector<std::vector<double>> Mt_aAt_a;
		Mt_aAt_a= MProduct(MTranspose(t_potential_a), MProduct(A, t_potential_a));
		double t_aAt_a= Mt_aAt_a[0][0];

		std::vector<std::vector<double>> Mt_bBt_b;
		Mt_bBt_b= MProduct(MTranspose(t_potential_b), MProduct(B, t_potential_b));
		double t_bBt_b= Mt_bBt_b[0][0];
		std::vector<std::vector<double>> Mt_bAt_b;
		Mt_bAt_b= MProduct(MTranspose(t_potential_b), MProduct(A, t_potential_b));
		double t_bAt_b= Mt_bAt_b[0][0];

		// AML cost for i'th data point
		cost_a = cost_a +  t_aAt_a/t_aBt_a ;
		cost_b = cost_b +  t_bAt_b/t_bBt_b ;
	}

	// Barrier term
	std::vector<std::vector<double>> Mt_aIt_a;
	Mt_aIt_a= MProduct(MTranspose(t_potential_a), MProduct(S.I, t_potential_a));
	double t_aIt_a= Mt_aIt_a[0][0];
	std::vector<std::vector<double>> Mt_aFt_a;
	Mt_aFt_a= MProduct(MTranspose(t_potential_a), MProduct(S.F, t_potential_a));
	double t_aFt_a= Mt_aFt_a[0][0];

	std::vector<std::vector<double>> Mt_bIt_b;
	Mt_bIt_b= MProduct(MTranspose(t_potential_b), MProduct(S.I, t_potential_b));
	double t_bIt_b= Mt_bIt_b[0][0];
	std::vector<std::vector<double>> Mt_bFt_b;
	Mt_bFt_b= MProduct(MTranspose(t_potential_b), MProduct(S.F, t_potential_b));
	double t_bFt_b= Mt_bFt_b[0][0];

	// add the barrier term
	cost_a = cost_a + pow((S.alpha*(t_aIt_a/t_aFt_a)), 2);
	cost_b = cost_b + pow((S.alpha*(t_bIt_b/t_bFt_b)), 2);

	// determine appropriate damping and if possible select an update
	if (cost_a >= current_cost && cost_b >= current_cost){
		// neither update reduced the cost
		S.theta_updated = false;
		// no change in the cost
		S.cost[0][S.k+1] = current_cost;
		// no change in parameters
		for(long i= 0; i<6; i++)
			S.t[i][S.k+1]= t[i][0];
		// no changes in step direction
		for(long i= 0; i<6; i++)
			S.delta[i][S.k+1]= delta[i][0];
		// next iteration add more Identity matrix
		S.lambda = S.lambda * S.damping_multiplier;
	}
	else if (cost_b < current_cost){
		// update 'b' reduced the cost function
		S.theta_updated = true;
		// store the new cost
		S.cost[0][S.k+1] = cost_b;
		// choose update 'b'
		double normt_b= norm_2(t_potential_b);
		for(long i= 0; i<6; i++)
			S.t[i][S.k+1]= t_potential_b[i][0]/normt_b;
		// store the step direction
		for(long i= 0; i<6; i++)
			S.delta[i][S.k+1]= update_b[i][0];
		// next iteration add less Identity matrix
		S.lambda = S.lambda / S.damping_multiplier;
	}
	else{
		// update 'a' reduced the cost function
		S.theta_updated = true;
		// store the new cost
		S.cost[0][S.k+1] = cost_a;
		// choose update 'a'
		double normt_a= norm_2(t_potential_a);
		for(long i= 0; i<6; i++)
			S.t[i][S.k+1]= t_potential_a[i][0]/normt_a;
		// store the step direction
		for(long i= 0; i<6; i++)
			S.delta[i][S.k+1]= update_a[i][0];
		// keep the same damping for the next iteration
		//S.lambda = S.lambda;
	}
}


void CFitEllipse::LineSearchStep(SzpakData &S)
//   Function: lineSearchStep
//
//   This function computes an update for the parameters
//   representing the ellipse, using the pseudo-inverse of a Gauss-Newton
//   approximation of the Hessian matrix. It then performs an inexact
//   line search to determine how far to move along the update direction
//   so that the resulting fit is still an ellipse
//
//   Parameters:
//      S a data structure containing various parameters
//        needed for the optimisation process
//
//   Returns: 
//     the same data structure S, except that relevant fields have been updated
//
//  Based on a matlab code by Zygmunt L. Szpak (c) 2012
{
	// extract local variables from data structure
	//--------------------------------------------
	std::vector<std::vector<double>> t(6, std::vector<double>(1, 0));
	for(long i= 0; i<6; i++)
		t[i][0]= S.t[i][S.k];
	std::vector<std::vector<double>> delta(6, std::vector<double>(1, 0));
	for(long i= 0; i<6; i++)
		delta[i][0]= S.delta[i][S.k];
	double current_cost = S.cost[0][S.k];

	// jacobian (vector), r(t)'d/dtheta[r(t)]
	std::vector<std::vector<double>> jacob;
	jacob= MProduct(MTranspose(MConcatenateRows(S.jacobian_matrix, S.jacobian_matrix_barrier)) , S.r);
	std::vector<std::vector<double>> MtFt;
	MtFt= MProduct(MTranspose(t), MProduct(S.F, t));
	double tFt= MtFt[0][0];

	// We solve for the new update direction in a numerically careful manner
	// If we didn't worry about numerical stability then we would compute 
	// the first new search direction like this:

	std::vector<std::vector<double>> update;
	// update = MScale(MProduct(MPseudoInverse(H), jacob), -1);

	// But close to the barrier between ellipses and hyperbolas we may
	// experience numerical conditioning problems due to the nature of the
	// barrier term itself.  Hence we perform the calculation in a
	// numerically more stable way with 

	std::vector<std::vector<double>> Z00;
	Z00= MAdd(MProduct(MTranspose(S.jacobian_matrix), S.jacobian_matrix), MScale(S.I, S.lambda));
	std::vector<std::vector<double>> Z01;
	Z01= MScale(MProduct(MTranspose(S.jacobian_matrix_barrier), S.jacobian_matrix_barrier), pow(tFt, 4));
	std::vector<std::vector<double>> Z10;
	Z10= S.I;
	std::vector<std::vector<double>> Z11;
	Z11= MScale(S.I, -pow(tFt, 4));
	std::vector<std::vector<double>> Z;
	Z= MConcatenateRows(MConcatenateColumns(Z00, Z01), MConcatenateColumns(Z10, Z11));

	std::vector<std::vector<double>> zz;
	std::vector<std::vector<double>> zeros(6, std::vector<double>(1, 0));
	zz= MScale(MConcatenateRows(jacob, zeros), -1);

	//Moore-Penrose inverse (pseudoinverse) of Z
	//update = pinv(Z,1e-20)*zz;
	update = MProduct(MPseudoInverse(Z, 1e-20), zz);
	// drop the nuisance parameter components
	update.resize(6);

	// there is no repeat...until construct so we use a while-do
	double frac = 0.5;
	double cost;
	std::vector<std::vector<double>> t_potential(6, std::vector<double>(1, 0));

	while (true){
		// compute potential update
		t_potential = MAdd(t, MScale(update, frac));
		delta = MScale(update, frac);
		// halve the step-size
		frac = frac/2.;
		// compute new residuals on data points
		cost = 0.;
		for(long i = 0; i<S.numberOfPoints; i++){
			// transformed data point
			std::vector<std::vector<double>> ux_j(6, std::vector<double>(1, 0));
			ux_j[0][0]= S.data_points[0][i]*S.data_points[0][i];
			ux_j[1][0]= S.data_points[0][i]*S.data_points[1][i];
			ux_j[2][0]= S.data_points[1][i]*S.data_points[1][i];
			ux_j[3][0]= S.data_points[0][i];
			ux_j[4][0]= S.data_points[1][i];
			ux_j[5][0]= 1.;

			// derivative of transformed data point
			std::vector<std::vector<double>> dux_j(6, std::vector<double>(2, 0));
			dux_j[0][0]= 2*S.data_points[0][i];
			dux_j[1][0]= S.data_points[1][i];
			dux_j[2][0]= 0.;
			dux_j[3][0]= 1.;
			dux_j[4][0]= 0.;
			dux_j[5][0]= 0.;
			dux_j[0][1]= 0;
			dux_j[1][1]= S.data_points[0][i];
			dux_j[2][1]= 2*S.data_points[1][i];
			dux_j[3][1]= 0.;
			dux_j[4][1]= 1.;
			dux_j[5][1]= 0.;

			// outer product
			std::vector<std::vector<double>> A;
			A = MProduct(ux_j, MTranspose(ux_j));

			// use identity covs
			std::vector<std::vector<double>> B;
			B = MProduct(dux_j, MTranspose(dux_j));

			std::vector<std::vector<double>> MtBt;
			MtBt= MProduct(MTranspose(t_potential), MProduct(B, t_potential));
			double tBt= MtBt[0][0];
			std::vector<std::vector<double>> MtAt;
			MtAt= MProduct(MTranspose(t_potential), MProduct(A, t_potential));
			double tAt= MtAt[0][0];

			// AML cost for i'th data point
			cost = cost + tAt/tBt;
		}

		// Barrier term
		std::vector<std::vector<double>> MtIt;
		MtIt= MProduct(MTranspose(t_potential), MProduct(S.I, t_potential));
		double tIt= MtIt[0][0];
		std::vector<std::vector<double>> MtFt;
		MtFt= MProduct(MTranspose(t_potential), MProduct(S.F, t_potential));
		double tFt= MtFt[0][0];

		// add the barrier term
		cost = cost + pow((S.alpha*(tIt/tFt)), 2);

		// check to see if cost function was sufficiently decreased, and whether
		// the estimate is still an ellipse. Additonally, if the step size
		// becomes too small we stop. 
		if ((tFt > 0 && (cost < (1-frac*S.gamma)*current_cost))  ||
			norm_2(delta) < S.tolDelta)
			break;
	}

	//Save output parameters
	S.theta_updated = true;
	double normt= norm_2(t_potential);
	for(long i= 0; i<6; i++)
		S.t[i][S.k+1]= t_potential[i][0]/normt;

	for(long i= 0; i<6; i++)
		S.delta[i][S.k+1]= delta[i][0];

	S.cost[0][S.k+1] = cost;
}


void CFitEllipse::guaranteedEllipseFit(long numStroke, long numSubStroke,
									   std::vector<double> &theta,
									   std::vector<std::vector<double>> pts)
//   Function: guaranteedEllipseFit
//
//   This function implements the ellipse fitting algorithm described in
//   Z.Szpak, W. Chojnacki and A. van den Hengel
//   "Guaranteed Ellipse Fitting with the Sampson Distance"
//   Proc. 12th European Conference on Computer Vision. ECCV 
//   Firenze,Italy, oct, 2012
//
//   Parameters:
//      theta an initial seed for parameters [a b c d e f] associated with
//            the equation   a x^2 + b x y + c y^2 + d x + e y + f = 0
//      pts   a 2xN matrix where N is the number of data points
//
//   Returns: 
//     a length-6 vector [a b c d e f] representing the parameters of the equation
//     a x^2 + b x y + c y^2 + d x + e y + f = 0
//     with the additional result that b^2 - 4 a c < 0.
//
//  Based on a matlab code by Zygmunt L. Szpak (c) 2012
{
	// ----------------------------------------------------------------
	// Open the Dialog that shows the progress, and allows cancellation
	//CDialogProgress Progress;
	//Progress.Create(IDD_DIALOG_Progress);
	//CTime HoraInicial= CTime::GetCurrentTime();

	//CString text;
	//if(numStroke < 0)
	//	text.Format("Fitting current elliptic arc by GEF...");
	//else if(numSubStroke > 0)
	//	text.Format("Fitting elliptic arc %i (subarc %i) by GEF...", numStroke, numSubStroke);
	//else
	//	text.Format("Fitting elliptic arc %i by GEF...", numStroke);
	//Progress.Message(text);
	//Progress.RangeProgress(0, 100);
	//Progress.UpdateProgress(0);
	// ----------------------------------------------------------------

	//Save input parameters vector as a local parameters matrix
	std::vector<std::vector<double>> t(6, std::vector<double>(1, 0));
	for(long i= 0; i<6; i++)
		t[i][0]= theta[i];

	// various variable initialisations
	//---------------------------------
	SzpakData S; //Define data structure

	// primary loop variable
	bool keep_going = true;
	// determines whether we use LineSearchStep or LevenbergMarquadtStep
	S.use_pseudoinverse = false;
	// in some case a LevenbergMarquardtStep does not decrease the cost
	// function and so the parameters (theta) are not updated
	S.theta_updated = false;
	// damping parameter in LevenbergMarquadtStep
	S.lambda = 0.01;
	// loop counter (matlab arrays start at index 1, not index 0)
	S.k = 0;
	// used to modify the tradeoff between gradient descent and hessian based
	// descent in LevenbergMarquadtStep
	S.damping_multiplier = 1.2;
	// used in LineSearchStep
	S.gamma = 0.00005;
	// number of data points
	S.numberOfPoints = pts[0].size();
	// barrier term that forces parameters to stay in elliptic region

	S.F.assign(6, std::vector<double>());
	for(long i= 0; i<6; i++)
		S.F[i].assign(6, 0.);
	S.F[0][0]= 0; S.F[0][1]= 0; S.F[0][2]= 2;
	S.F[1][0]= 0; S.F[1][1]=-1; S.F[1][2]= 0;
	S.F[2][0]= 2; S.F[2][1]= 0; S.F[2][2]= 0;

	S.I.assign(6, std::vector<double>());
	for(long i= 0; i<6; i++){
		S.I[i].assign(6, 0.);
		S.I[i][i]= 1.;
	}

	// homotopy parameters that weighs contribution of barrier term
	S.alpha = 1e-3;
	// data points that we are going to fit an ellipse to
	S.data_points = pts;

	// various parameters that determine stopping criteria
	//----------------------------------------------------
	// maximum loop iterations 
	long maxIter = 100;
	// step-size tolerance
	S.tolDelta = 1e-7;
	// cost tolerance
	S.tolCost = 1e-7;
	// parameter tolerance
	S.tolTheta = 1e-7;

	// various initial memory allocations
	//-----------------------------------
	// allocate space for cost of each iteration
	S.cost.assign(1, std::vector<double>());
	for(long i= 0; i<1; i++)
		S.cost[i].assign(maxIter, 0.);

	// allocate space for the parameters of each iteration
	S.t.assign(6, std::vector<double>());
	for(long i= 0; i<6; i++)
		S.t[i].assign(maxIter, 0.);

	// allocate space for the parameter direction of each iteration
	S.delta.assign(6, std::vector<double>());
	for(long i= 0; i<6; i++)
		S.delta[i].assign(maxIter, 0.);

	// make parameter vector a unit norm vector for numerical stability
	double normt= norm_2(MTranspose(t)[0]);  //Argument must be a vector
	for(long i= 0; i<6; i++)
		t[i][0] = t[i][0]/normt;

	//////// store the parameters associated with the first iteration
	for(long i= 0; i<6; i++)
		S.t[i][S.k]= t[i][0];

	// start with some random search direction (here we choose all 1's)
	// we can initialise with anything we want, so long as the norm of the
	// vector is not smaller than tolDeta. The initial search direction 
	// is not used in any way in the algorithm. 
	for(long i= 0; i<6; i++)
		S.delta[i][S.k]= 1.;

	// main estimation loop
	bool State= true;
	while (State && keep_going && S.k < maxIter-1){
		// allocate space for residuals (+1 to store barrier term) 
		S.r.assign(S.numberOfPoints+1, std::vector<double>());
		for(long i= 0; i<S.numberOfPoints+1; i++)
			S.r[i].assign(1, 0.);

		// allocate space for the jacobian matrix based on AML component
		S.jacobian_matrix.assign(S.numberOfPoints, std::vector<double>());
		for(long i= 0; i<S.numberOfPoints; i++)
			S.jacobian_matrix[i].assign(6, 0.);

		// allocate space for the jacobian matrix based on Barrier component
		S.jacobian_matrix_barrier.assign(1, std::vector<double>());
		for(long i= 0; i<1; i++)
			S.jacobian_matrix_barrier[i].assign(6, 0.);

		// grab the current parameter estimates
		for(long i= 0; i<6; i++)
			t[i][0]= S.t[i][S.k];

		// residuals computed on data points
		for(long i = 0; i<S.numberOfPoints; i++){
			// transformed data point
			std::vector<std::vector<double>> ux_j(6, std::vector<double>(1, 0));
			ux_j[0][0]= S.data_points[0][i]*S.data_points[0][i];
			ux_j[1][0]= S.data_points[0][i]*S.data_points[1][i];
			ux_j[2][0]= S.data_points[1][i]*S.data_points[1][i];
			ux_j[3][0]= S.data_points[0][i];
			ux_j[4][0]= S.data_points[1][i];
			ux_j[5][0]= 1.;

			// derivative of transformed data point
			std::vector<std::vector<double>> dux_j(6, std::vector<double>(2, 0));
			dux_j[0][0]= 2*S.data_points[0][i];
			dux_j[1][0]= S.data_points[1][i];
			dux_j[2][0]= 0.;
			dux_j[3][0]= 1.;
			dux_j[4][0]= 0.;
			dux_j[5][0]= 0.;
			dux_j[0][1]= 0;
			dux_j[1][1]= S.data_points[0][i];
			dux_j[2][1]= 2*S.data_points[1][i];
			dux_j[3][1]= 0.;
			dux_j[4][1]= 1.;
			dux_j[5][1]= 0.;

			// outer product
			std::vector<std::vector<double>> A(6, std::vector<double>(6, 0));
			A= MProduct(ux_j, MTranspose(ux_j));

			// use identity covs
			std::vector<std::vector<double>> B(6, std::vector<double>(6, 0));
			B= MProduct(dux_j, MTranspose(dux_j));

			std::vector<std::vector<double>> MtBt;
			MtBt= MProduct(MTranspose(t), MProduct(B, t));
			double tBt= MtBt[0][0];
			std::vector<std::vector<double>> MtAt;
			MtAt= MProduct(MTranspose(t), MProduct(A, t));
			double tAt= MtAt[0][0];

			// AML cost for i'th data point
			S.r[i][0] = sqrt(tAt/tBt);

			// derivative AML component
			std::vector<std::vector<double>> M;
			M= MScale(A, 1./tBt);
			std::vector<std::vector<double>> Xbits;
			Xbits= MScale(B, tAt/(tBt*tBt));
			std::vector<std::vector<double>> X;
			X= MSubtract(M, Xbits);

			// gradient for AML cost function 
			std::vector<std::vector<double>> grad;
			grad= MScale(MProduct(X, t), 1./sqrt(tAt/tBt));
			// build up jacobian matrix
			//S.jacobian_matrix[i]= grad;
			for(long j= 0; j<6; j++)
				S.jacobian_matrix[i][j]= grad[j][0];
		}

		// Barrier term
		std::vector<std::vector<double>> MtIt;
		MtIt= MProduct(MTranspose(t), MProduct(S.I, t));
		double tIt= MtIt[0][0];
		std::vector<std::vector<double>> MtFt;
		MtFt= MProduct(MTranspose(t), MProduct(S.F, t));
		double tFt= MtFt[0][0];

		// add the penalty term
		S.r[S.numberOfPoints][0] = S.alpha*(tIt/tFt);

		// Derivative barrier component
		std::vector<std::vector<double>> N;
		N = MScale(S.I, 1./tFt);
		std::vector<std::vector<double>> Ybits;
		Ybits = MScale(S.F, tIt/(tFt*tFt));
		std::vector<std::vector<double>> Y;
		Y = MSubtract(N, Ybits);
		std::vector<std::vector<double>> grad_penalty;
		grad_penalty= MScale(MProduct(Y, t), 2*S.alpha);
		for(long j= 0; j<6; j++)
			S.jacobian_matrix_barrier[0][j]= grad_penalty[j][0];

		// Jacobian matrix after combining AML and barrier terms
		S.jacobian_matrix_full = MConcatenateRows(S.jacobian_matrix, S.jacobian_matrix_barrier);

		// approximate Hessian matrix
		S.H = MProduct(MTranspose(S.jacobian_matrix_full), S.jacobian_matrix_full);

		// sum of squares cost for the current iteration
		//struct.cost(k) = 0.5*(struct.r'*struct.r);
		std::vector<std::vector<double>> Mcost;
		Mcost = MProduct(MTranspose(S.r), S.r);
		S.cost[0][S.k] = Mcost[0][0];

		// If we haven't overshot the barrier term
		// then we use LevenbergMarquadt step
		if (!S.use_pseudoinverse){
			LevenbergMarquardtStep(S);
		}
		else{
			LineSearchStep(S);
		}

		// Check if the latest update overshot the barrier term 
		std::vector<std::vector<double>> tk1(6, std::vector<double>(6, 0));
		for(long i= 0; i<(long)t.size(); i++)
			tk1[i][0]= S.t[i][S.k+1];
		//std::vector<std::vector<double>> MtFt;
		MtFt= MProduct(MTranspose(tk1), MProduct(S.F, tk1));

		if (MtFt[0][0] <= 0){
			// from now onwards we will only use lineSearchStep to ensure
			// that we do not overshoot the barrier 
			S.use_pseudoinverse = true;
			S.lambda = 0;

			//Revert to previous iterate which was an ellipse
			for(long i= 0; i<6; i++)
				S.t[i][S.k+1] = S.t[i][S.k];

			if (S.k > 1){
				for(long i= 0; i<6; i++)
					S.t[i][S.k] = S.t[i][S.k-1];
			}
		}
		else{
			// Check for various stopping criteria to end the main loop
			std::vector<double> tkp(6, 0);
			for(long i= 0; i<6; i++)
				tkp[i] = S.t[i][S.k+1] + S.t[i][S.k];
			double n1= norm_2(tkp);

			std::vector<double> tkm(6, 0);
			for(long i= 0; i<6; i++)
				tkm[i] = S.t[i][S.k+1] - S.t[i][S.k];
			double n2= norm_2(tkm);

			std::vector<double> d(6, 0);
			for(long i= 0; i<6; i++)
				d[i] = S.delta[i][S.k+1];
			double n3= norm_2(d);

			if (std::min(n1, n2) < S.tolTheta && S.theta_updated)
				keep_going = false;
			else if (abs(S.cost[0][S.k] - S.cost[0][S.k+1]) < S.tolCost && S.theta_updated)
				keep_going = false;
			else if (n3 < S.tolDelta && S.theta_updated)
				keep_going = false;
		}

		S.k = S.k + 1;

		// ----------------------------------------------------------------
		//State= Progress.UpdateProgress((long)(100.*S.k/(double)maxIter));
		// ----------------------------------------------------------------
	}

	//Output current theta
	for(long i= 0; i<6; i++)
		theta[i] = S.t[i][S.k];

	double normtheta= norm_2(theta);
	for(long i= 0; i<6; i++)
		theta[i] = theta[i]/normtheta;

	// ----------------------------------------------------------------
	//Progress.DestroyWindow();
	// ----------------------------------------------------------------
}


bool CFitEllipse::FitEllipseGEF(CStroke InputStroke, long numStroke, long numSubStroke, CConicParams &phi)
//   Function: compute_guaranteedellipse_estimates

//   This function is a wrapper for an approximate maximum likelihood guaranteed
//   ellipse fit due to

//   Z.Szpak, W. Chojnacki and A. van den Hengel
//   "Guaranteed Ellipse Fitting with the Sampson Distance"
//   Proc. 12th European Conference on Computer Vision. ECCV 
//   Firenze,Italy, oct, 2012

//   It first shifts all the data points to a new coordinate system so that
//   the origin of the coordinate system is at the center of the data points, 
//   and then scales all data points so that they lie more or less within
//   a unit box.
//   It is within this transformed coordinate system that the 
//   ellipse is estimated.
//   Since the approximate maximum likelihood method
//   is an iterative scheme, it requires an initial seed for the parameters.
//   The seed is taken to be the direct ellipse fit due to 

//   R. Halif and J. Flusser
//   "Numerically stable direct least squares fitting of ellipses"
//   Proc. 6th International Conference in Central Europe on Computer Graphics and Visualization. WSCG '98
//   Czech Republic,125--132, feb, 1998
//
//   Their work is a numerically stable version of

//   A. W. Fitzgibbon, M. Pilu, R. B. Fisher
//   "Direct Least Squares Fitting of Ellipses"
//   IEEE Trans. PAMI, Vol. 21, pages 476-480 (1999)

//   The final ellipse parameters estimates are transformed 
//   back to the original data space. 

//   Parameters:
//      pts  a 3xN matrix where N is the number of data points

//   Returns: 
//     a length-6 vector [a b c d e f] representing the parameters of the equation
//     a x^2 + b x y + c y^2 + d x + e y + f = 0
//     with the additional result that b^2 - 4 a c < 0.

//  Based on a matlab code by Zygmunt L. Szpak (c) 2012
{
	// Data points in homogeneous coordinates
	long n = InputStroke.GetNumPoints();
	std::vector<std::vector<double>> pts(3, std::vector<double>(n, 0));

	for (long i=0; i<n; i++){
		POINT2D pp= InputStroke.GetPoint(i).GetPosition();

		pts[0][i] = pp.x;
		pts[1][i] = pp.y;
		pts[2][i] = 1.0;
	}

	// hartley normalise the data
	std::vector<std::vector<double>> T(3, std::vector<double>(3, 0));
	normalise2dpts(pts, T);

	pts= MProduct(T, pts);

	//Seed parameters
	//DIR fit should not be re-calculated here,
	//since the input parameters already contain a DIR fit
	//But this approach requires normalised data
	CStroke StrokeNorm= InputStroke;
	for (long i=0; i<n; i++){
		CStrokePoint pp;

		pp.PutPosition(pts[0][i], pts[1][i]);
		pp.PutTime(InputStroke.GetPoint(i).GetTime());
		StrokeNorm.PutPoint(i, pp);
	}

	if(!FitEllipseDIR(StrokeNorm, phi)){
		if(FitEllipse5P(StrokeNorm, phi))
			phi.b= -phi.b;
		else{
			return false;
		}
	}

	std::vector<double> theta(6,0);
	phi.GetParams(theta[0], theta[1], theta[2], theta[3], theta[4], theta[5]);

	double normtheta= norm_2(theta);
	for(long i= 0; i<6; i++)
		theta[i] = theta[i]/normtheta;

	//Fitting
	//WARNING: Time consuming tasks are in the main loop of guaranteedEllipseFit!
	guaranteedEllipseFit(numStroke, numSubStroke, theta, pts);

	//Denormalise C
	std::vector<std::vector<double>> C(3, std::vector<double>(3, 0));
	C[0][0] = theta[0];      C[0][1] = theta[1]/2.;   C[0][2] = theta[3]/2.;
	C[1][0] = theta[1]/2.;   C[1][1] = theta[2];      C[1][2] = theta[4]/2.;
	C[2][0] = theta[3]/2.;   C[2][1] = theta[4]/2.;   C[2][2] = theta[5];

	C= MProduct(MTranspose(T), MProduct(C, T));

	theta[0] = C[0][0];
	theta[1] = C[0][1]*2;
	theta[2] = C[1][1];
	theta[3] = C[0][2]*2;
	theta[4] = C[1][2]*2;
	theta[5] = C[2][2];
	normtheta= norm_2(theta);
	for(long i= 0; i<6; i++)
		theta[i] = theta[i]/normtheta;

	phi.PutParams(theta[0], theta[1], theta[2], theta[3], theta[4], theta[5]);

	return true;
}


double CFitEllipse::DistancePointEllipse(const double e[2], const double y[2], double x[2])
// The function returns the distance from
// the query point to the ellipse.
// It also computes the ellipse point (x0,x1)
// in the first quadrant that is closest to (y0,y1).

// The ellipse is (x0/e0)^2 + (x1/e1)^2 = 1 with e0 >= e1.
// The query point is (y0,y1) with y0 >= 0 and y1 >= 0. 
{
	double distance= 0.;

	if (y[1] > 0.){
		if (y[0] > 0.){
			// Bisect to compute the root of F(t) for t >= -e1*e1.
			double esqr[2] = { e[0]*e[0], e[1]*e[1] };
			double ey[2] = { e[0]*y[0], e[1]*y[1] };
			double t0 = -esqr[1] + ey[1];
			double t1 = -esqr[1] + sqrt(ey[0]*ey[0] + ey[1]*ey[1]);
			double t = t0;
			const int imax = 2*std::numeric_limits<double>::max_exponent;
			for (int i = 0; i < imax; ++i){
				t = 0.5*(t0 + t1);
				if (t == t0 || t == t1)
					break;

				double r[2] = { ey[0]/(t + esqr[0]), ey[1]/(t + esqr[1]) };
				double f = r[0]*r[0] + r[1]*r[1] - 1.;
				if (f > 0.)
					t0 = t;
				else if (f < 0.)
					t1 = t;
				else
					break;
			}

			x[0] = esqr[0]*y[0]/(t + esqr[0]);
			x[1] = esqr[1]*y[1]/(t + esqr[1]);
			double d[2] = { x[0] - y[0], x[1] - y[1] };
			distance = sqrt(d[0]*d[0] + d[1]*d[1]);
		}
		else{ // y0 == 0
			x[0] = 0.;
			x[1] = e[1];
			distance = fabs(y[1] - e[1]);
		}
	}
	else{ // y1 == 0
		double denom0 = e[0]*e[0] - e[1]*e[1];
		double e0y0 = e[0]*y[0];
		if (e0y0 < denom0){
			// y0 is inside the subinterval.
			double x0de0 = e0y0/denom0;
			double x0de0sqr = x0de0*x0de0;
			x[0] = e[0]*x0de0;
			x[1] = e[1]*sqrt(fabs(1. - x0de0sqr));
			double d0 = x[0] - y[0];
			distance = sqrt(d0*d0 + x[1]*x[1]);
		}
		else{
			// y0 is outside the subinterval. The closest ellipse point has
			// x1 == 0 and is on the domain-boundary interval (x0/e0)^2 = 1.
			x[0] = e[0];
			x[1] = 0.;
			distance = fabs(y[0] - e[0]);
		}
	}
	return distance;
}


POINT2D CFitEllipse::ClosestPointInEllipse(double Ra, double Rb, POINT2D P)
//The function finds the point of the ellipse (defined by its radius Ra and Rb),
//which is closest to the point defined by P

//The ellipse is centred in the origin and with its main axis horizontal

//The distance between some point and an ellipse is a transcendental equation
//Root finding techniques are required

// The function returns the distance from the query point to the ellipse.
// It also computes the ellipse point (x0,x1) that is closest to (y0,y1).
//(This point is the one whose normal passes through P)

//This function reflects components of (y0; y1) as needed
//to place the query point in the first quadrant,
//sorts the (e0; e1) values to be correctly ordered,
//and then calls DistancePointEllipseSpecial.
//The returned (x0; x1) is postprocessed
//to match the quadrant of the query point.

// The ellipse is (x0/e0)^2 + (x1/e1)^2 = 1. The query point is (y0,y1).

//This function is a modified version of "DistancePointEllipse", provided by:
	// Geometric Tools, LLC
	// Copyright (c) 1998-2013
	// Distributed under the Boost Software License, Version 1.0.
	// http://www.boost.org/LICENSE_1_0.txt
	// http://www.geometrictools.com/License/Boost/LICENSE_1_0.txt
	//
	// File Version: 5.6.3 (2013/01/03)

	//D. Eberly.
	//Internet publication: "Distance from a point to an ellipse in 2D" (2004) 
	//Geometric Tools, LLC, www.geometrictools.com
	//Book publication: "3D Game Engine Design", 2nd edition.
	//Morgan Kaufmann Publishers, San Francisco, CA, 2007.
	//(see Section 14.13.1)
{
	//Save data in internal format
	double e[2];
	e[0]= Ra;
	e[1]= Rb;

	double y[2];
	y[0]= P.x;
	y[1]= P.y;

	//Determine reflections for y to the first quadrant
	bool reflect[2];
	for (long i = 0; i < 2; ++i){
		reflect[i] = (y[i] < 0.);
	}

	//Determine the axis order for decreasing extents
	int permute[2];
	if (e[0] < e[1]){
		permute[0] = 1; permute[1] = 0;
	}
	else{
		permute[0] = 0; permute[1] = 1;
	}

	int invpermute[2];
	for (long i = 0; i < 2; ++i){
		invpermute[permute[i]] = i;
	}

	double locE[2], locY[2];
	for (long i = 0; i < 2; ++i){
		long j = permute[i];
		locE[i] = e[j];
		locY[i] = y[j];
		if (reflect[j]){
			locY[i] = -locY[i];
		}
	}

	double locX[2];
	double distance = DistancePointEllipse(locE, locY, locX);

	//Restore the axis order and reflections
	double x[2];
	for (long i = 0; i < 2; ++i){
		long j = invpermute[i];
		if (reflect[i]){
			locX[j] = -locX[j];
		}
		x[i] = locX[j];
	}

	POINT2D T(x[0], x[1]);

	return T;
}


void CFitEllipse::GetEndPoints(CStroke Stroke, CArc &Arc, bool EndPointsFitted)
//The function finds the points of the ellipse
//that are closer to the endpoints of the stroke

//The function also discriminates the arc that best fits the stroke

//Adds the polar angles of both endpoints into Arc in counterclock sequence
//	StartAngle < EndAngle

//WARNING! All those parameters are calculated in the "ellipse system"
{
	//Get the stroke endpoints and the midpoint
	POINT2D P0= Stroke.GetPoint(0).GetPosition();
	POINT2D P1= Stroke.GetPoint(Stroke.GetNumPoints()-1).GetPosition();
	POINT2D PM= Stroke.GetPoint(Stroke.GetNumPoints()/2).GetPosition();

	//Transform to "ellipse system"
	POINT2D Centre(Arc.Centre_x, Arc.Centre_y);
	P0= P0 - Centre;
	P1= P1 - Centre;
	P0= P0.Rotate(-Arc.Rotation);
	P1= P1.Rotate(-Arc.Rotation);

	PM= PM - Centre;
	PM= PM.Rotate(-Arc.Rotation);

	//Find the closest points in the ellipse
	POINT2D T0, T1;
	if(!EndPointsFitted){
		T0= ClosestPointInEllipse(Arc.Radius_a, Arc.Radius_b, P0);
		T1= ClosestPointInEllipse(Arc.Radius_a, Arc.Radius_b, P1);
	}
	else{
		T0= P0;
		T1= P1;
	}
	Arc.GapEnds= std::max(MODULE_2D(T0, P0), MODULE_2D(T1, P1));

	POINT2D TM= ClosestPointInEllipse(Arc.Radius_a, Arc.Radius_b, PM);

	//Calculate their polar angles
	double A0= acos(T0.x/Arc.Radius_a);
	if(T0.y<0) A0= 2*NUMBER_PI-A0;
	double A1= acos(T1.x/Arc.Radius_a);
	if(T1.y<0) A1= 2*NUMBER_PI-A1;
	double AM= acos(TM.x/Arc.Radius_a);
	if(TM.y<0) AM= 2*NUMBER_PI-AM;

	//Sort from minor to major angle
	double AA0= std::min(A0, A1);
	double AA1= std::max(A0, A1);

	if(AA0 < AM && AM < AA1){
		Arc.StartAngle= AA0; //Start angle is in the range [0, 2*PI)
		Arc.EndAngle= AA1;
	}
	else{
		Arc.StartAngle= AA1;
		Arc.EndAngle= AA0 + 2*NUMBER_PI;  //End angle is in the range (0, 4*PI]
	}

	Arc.StartAngle+= Arc.Rotation;
	Arc.EndAngle+= Arc.Rotation;
	if(Arc.StartAngle < 0){
		Arc.StartAngle+= 2*NUMBER_PI;
		Arc.EndAngle+= 2*NUMBER_PI;
	}

	/////////////////////////////////////////////////////////
	//Save to display
	//T0= T0.Rotate(Arc.Rotation) + Centre;
	//TM= TM.Rotate(Arc.Rotation) + Centre;
	//T1= T1.Rotate(Arc.Rotation) + Centre;

	//Arc.T0_x= T0.x;
	//Arc.T0_y= T0.y;
	//Arc.TM_x= TM.x;
	//Arc.TM_y= TM.y;
	//Arc.T1_x= T1.x;
	//Arc.T1_y= T1.y;
	/////////////////////////////////////////////////////////
}


bool CFitEllipse::GetArc(CConicParams phi, CStroke Stroke,CArc &Arc, bool EndPointsFitted)
//The input is the general quadratic curve ax^2+bxy+cy^2+dx+ey+f=0
//intended at representing an ellipse

//The output is the canonical form of the elliptic arc, which includes:
//	the centre (Arc.Centre_x, Arc.Centre_y)
//	the semi-axes lengths (Arc.Radius_a, Arc.Radius_b)
//	the counterclockwise angle of rotation from the x-axis to the major axis (Arc.Rotation)
//	The foci (Arc.Focus1_x, Arc.Focus1_y), (Arc.Focus2_x, Arc.Focus2_y)
//	The endpoints (the points of the ellipse closer to the endpoints of the stroke)

//Returns false if any parameter is invalid
{
	//Center of the ellipse
	//---------------------
	double J = phi.a*phi.c - phi.b*phi.b/4; //|a b/2; b/2 c|

	Arc.Centre_x = (phi.b*phi.e-2*phi.c*phi.d)/(4*J);
	Arc.Centre_y = (phi.b*phi.d-2*phi.a*phi.e)/(4*J);

	//Radius and rotation
	//-------------------
	if(phi.a==phi.c && phi.b==0.){
		//Is a circle
		Arc.Radius_a = sqrt((phi.d*phi.d+phi.e*phi.e)/(4*phi.a*phi.a) - phi.f/phi.a);
		Arc.Radius_b = Arc.Radius_a;
		Arc.Rotation= 0.;
	}
	else{
		//Semi-axes lengths
		double aa1= (phi.a*phi.e*phi.e+phi.c*phi.d*phi.d+phi.f*phi.b*phi.b-phi.b*phi.d*phi.e)/2. - 2*phi.a*phi.c*phi.f;
		double aa3= sqrt((phi.a-phi.c)*(phi.a-phi.c)+phi.b*phi.b);

		Arc.Radius_a = sqrt(aa1/(J*(phi.a+phi.c-aa3)));
		Arc.Radius_b = sqrt(aa1/(J*(phi.a+phi.c+aa3)));

		//Counterclockwise angle of rotation from the x-axis to the major axis of the ellipse
		if(phi.b==0){
			if(phi.a<phi.c)
				Arc.Rotation= 0.;
			else
				Arc.Rotation= NUMBER_PI/2;
		}
		else{
			if(phi.a<phi.c)
				Arc.Rotation= atan(phi.b/(phi.a-phi.c))/2;
			else
				Arc.Rotation= NUMBER_PI/2 + atan(phi.b/(phi.a-phi.c))/2;
		}

		//Arc.Rotation= atan2(phi.b, (phi.a-phi.c))/2;
		if(Arc.Rotation < 0) Arc.Rotation+= NUMBER_PI;
	}

	//Check validity of the canonical form
	if(_isnan((float)Arc.Radius_a)  ||
	   _isnan((float)Arc.Radius_b) ||
	   _isnan((float)Arc.Rotation))
		return false;

	//Foci
	//----
	double cc= sqrt(Arc.Radius_a*Arc.Radius_a - Arc.Radius_b*Arc.Radius_b);
	double ccx= cc * cos(Arc.Rotation);
	double ccy= cc * sin(Arc.Rotation);
	Arc.Focus1_x= Arc.Centre_x + ccx;
	Arc.Focus1_y= Arc.Centre_y + ccy;
	Arc.Focus2_x= Arc.Centre_x - ccx;
	Arc.Focus2_y= Arc.Centre_y - ccy;

	//Endpoints
	//---------
	GetEndPoints(Stroke, Arc, EndPointsFitted);

	return true;
}


void CFitEllipse::GetStrokeMetrics(CStroke Stroke, CStrokeMetrics &SM)
//Returns metrics of the troke

//		BoxLength measures the length of the minimum box
//		that contains the stroke

//		Sagitta measures the width of the minimum box
//		that contains the stroke

//		ShortStroke is set to true if the stroke defines a "short" elliptic arc
//		(arc encompassing an angle shorter than 180 degrees)
//		This indirectly measured through the "Gap", which is the maximum distance between
//		every endpoint and the closest side of the bounding box

//		FlatStroke is set to true if the curvature is small (the arc is "flat")
//		Stroke is considered flat if the ratio between its sagitta and its chord
//		is less than FLATNESS

//		SmallAndFlat is set to true if ShortStroke and FlatStroke are true,
//		and false otherwise

//Those metrics are used to check if the arc must be fitted again
{
	//Rotate the stroke
	//-----------------
	CStroke RotatedStroke= Stroke;
	RotatedStroke.Rotate(Stroke.GetCentroid(), -Stroke.GetSlope180());

	//Stroke metrics
	//--------------
	RECTANGLE2D bounds= RotatedStroke.GetBounds();
	SM.BoxLength= abs(bounds.right-bounds.left);

	SM.Sagitta= bounds.top-bounds.bottom;
	POINT2D F= RotatedStroke.GetPoint(0).GetPosition();
	POINT2D L= RotatedStroke.GetPoint(RotatedStroke.GetNumPoints()-1).GetPosition();

	double Chord= MODULE_2D(F, L);
	double FLATNESS= 0.2;
	bool FlatStroke= (SM.Sagitta < FLATNESS*Chord);

	double Gap= std::max(abs(std::min(F.x, L.x) - bounds.left), abs(bounds.right - std::max(F.x, L.x)));
	double TOLERANCE= 0.1*std::max(Chord/2., SM.Sagitta);
	SM.ShortStroke= (Gap < TOLERANCE);

	SM.SmallAndFlat= (SM.ShortStroke && FlatStroke);
}


bool CFitEllipse::GetConvexHull(CStroke &Stroke)
//Calculates the Convex Hull of the Stroke

//Returns true if the resulting stroke still may depict an elliptic arc
//(i.e. contains five or more points)
{
	CStroke S= Stroke;
	S.GetConvexHull();
	if(S.GetNumPoints() < 5)
		//Convex Hull is not a valid stroke
		return false;

	Stroke= S;
	return true;
}


bool CFitEllipse::IsSmallArc(CArc Arc)
//Returns true if Arc encompasses and angle smaller than 180 degrees
{
	if(abs(Arc.EndAngle-Arc.StartAngle) < NUMBER_PI)
		return true;
	else
		return false;
}


bool CFitEllipse::IsArcOverflow(CStroke Stroke, CArc Arc,
								double BoxLength, double Sagitta)
//Returns true if the arc exceeds (overflows) the bounding box of the stroke

//WARNING! This flag is only valid if the stoke depicts a long arc
{
	double TOLOVERFLOW= 1.15; //Tolerance is set to 15% of the size of the bounding box

	double StrokeRotation= Stroke.GetSlope180();
	double delta= abs(StrokeRotation-Arc.Rotation);

	if((delta < NUMBER_PI/8. || delta > 7./8.*NUMBER_PI)){  //Parallel
		if((2*Arc.Radius_a > TOLOVERFLOW*BoxLength) ||
		    (Arc.Radius_b > TOLOVERFLOW*Sagitta))  //Only valid for long arcs
			return true;
	}

	else if((delta > 3./8.*NUMBER_PI && delta < 5./8.*NUMBER_PI)){  //Perpendicular
		if((2*Arc.Radius_b > TOLOVERFLOW*BoxLength) ||
		    (Arc.Radius_a > TOLOVERFLOW*Sagitta))  //Only valid for long arcs
			return true;
	}

	//else if(Arc.Radius_a > sqrt(BoxLength*BoxLength + Sagitta*Sagitta))
	else if(sqrt(Arc.Radius_a*Arc.Radius_a + Arc.Radius_b*Arc.Radius_b) >
			sqrt(BoxLength*BoxLength/4 + Sagitta*Sagitta)*TOLOVERFLOW)
		return true;

	return false;
}


bool CFitEllipse::IsGapEnds(CStroke Stroke, CArc Arc)
//Returns true if at least one of the endpoints of the stroke is too distant from the arc
{
	double MAXGAP= 0.05; //Gap threshold (% of the length)

	if(Arc.GapEnds > Stroke.GetLength(10)*MAXGAP)
		return true;

	return false;
}


bool CFitEllipse::IsArcUnderflow(CStroke Stroke, CArc Arc,
								double BoxLength, double Sagitta)
//Returns true if the stroke exceeds the bounding box of the arc
//(Detects the bias of DIR approach towards smaller ellipses)

//WARNING! This flag is only valid if the stoke depicts a short arc
{
	double TOLUNDERFLOW= 0.85;  //Tolerance is set to 15% of the size of the bounding box

	double StrokeRotation= Stroke.GetSlope180();
	double delta= abs(StrokeRotation-Arc.Rotation);

	if((delta < NUMBER_PI/8. || delta > 7./8.*NUMBER_PI)){  //Parallel
		if((2*Arc.Radius_a < TOLUNDERFLOW*BoxLength) ||
		    (Arc.Radius_b < TOLUNDERFLOW*Sagitta))  //Only valid for short arcs
			return true;
	}

	else if((delta > 3./8.*NUMBER_PI && delta < 5./8.*NUMBER_PI)){  //Perpendicular
		if((2*Arc.Radius_b < TOLUNDERFLOW*BoxLength) ||
		   (Arc.Radius_a < TOLUNDERFLOW*Sagitta))  //Only valid for short arcs
			return true;
	}

	//else if(Arc.Radius_a < sqrt(BoxLength*BoxLength + Sagitta*Sagitta))
	else if(sqrt(Arc.Radius_a*Arc.Radius_a + Arc.Radius_b*Arc.Radius_b) <
			sqrt(BoxLength*BoxLength/4 + Sagitta*Sagitta)*TOLUNDERFLOW)
		return true;

	return false;
}


bool CFitEllipse::IsBadFit5P(CStroke Stroke, CArc Arc, CStrokeMetrics SM)
//Returns true if the stroke seems to fit well, and false otherwise
{
	bool SmallArc= IsSmallArc(Arc);

	return((SM.SmallAndFlat && !SmallArc) ||
	       (!SM.ShortStroke && SmallArc) ||
	     //(SM.ShortStroke && !SmallArc) ||
	       (!SmallArc && IsArcOverflow(Stroke, Arc, SM.BoxLength, SM.Sagitta)));
}


bool CFitEllipse::Fit5P(CStroke Stroke, CArc &Arc,
						bool UseConvexHull5P, bool CheckQuality, CStrokeMetrics SM)
//Returns the parameters of the ellipse arc that best fit the Stroke
//Uses the approach Five-points applied to the convex hull of the stroke (5P/CH)
//Alternatively uses 5P applied to the original stroke
//Returns false if a valid fit cannnot be found

//If flag CheckQuality is true, the function checks to detect poor quality fits
//If so, tries to fit its convex hull,
//and returns false if such poor quality fits persist

//5P is considerd poor if:
//	The stroke is short and flat, and the arc is not short (does not cover a small angle)
//	The arc is greater than (overflows) the bounding box of the stroke
//	At least one of the endpoints of the stroke is too distant from the arc
//	The arc is smaller than (underflows) the boundig box
{
	//Get a 5P fit
	bool ConvexHull;
	CConicParams phi;

	CStroke SCH= Stroke;
	if(UseConvexHull5P && GetConvexHull(SCH)){
		//Try first 5P/CH and second 5P
		if(FitEllipse5P(SCH, phi))
			ConvexHull= true;
		else if(FitEllipseSpeed==0 && FitEllipse5P(Stroke, phi))
			ConvexHull= false;
		else
			return false;
	}
	else{
		//Try directly 5P
		if(FitEllipse5P(Stroke, phi))
			ConvexHull= false;
		else
			return false;
	}

	if(!GetArc(phi, Stroke, Arc, true)){
		if(!ConvexHull)
			return false;
		else 
			if(FitEllipse5P(Stroke, phi) && GetArc(phi, Stroke, Arc, true))
				ConvexHull= false;
			else
				return false;
	}

	//Check if fit is poor
	if(CheckQuality){
		if(DisplayColours){
			Arc.ColorR= 255; Arc.ColorG= 0; Arc.ColorB= 0;
		}

		if(IsBadFit5P(Stroke, Arc, SM)){
			if(ConvexHull){
				//Try with 5P
				if(FitEllipse5P(Stroke, phi) &&
				   GetArc(phi, Stroke, Arc, true)){
					ConvexHull= false;
					if(IsBadFit5P(Stroke, Arc, SM))
						return false;
				}
				else
					return false;
			}
			else
				return false;
		}
	}

	//Assign colour to display the arc
	if(DisplayColours){
		if(ConvexHull){
			Arc.ColorR= 0; Arc.ColorG= 255; Arc.ColorB= 155;
		}
		else{
			Arc.ColorR= 0; Arc.ColorG= 155; Arc.ColorB= 0;
		}
	}

	return true;
}


bool CFitEllipse::IsBadFitDIR(CStroke Stroke, CArc Arc, CStrokeMetrics SM)
//Returns true if the stroke seems to fit well, and false otherwise
{
	bool SmallArc= IsSmallArc(Arc);

	return((SM.SmallAndFlat && !SmallArc) ||
		   (!SM.ShortStroke && SmallArc) ||
		   IsGapEnds(Stroke, Arc) ||
		   (SM.SmallAndFlat && IsArcUnderflow(Stroke, Arc, SM.BoxLength, SM.Sagitta)) ||
		   (!SmallArc && IsArcOverflow(Stroke, Arc, SM.BoxLength, SM.Sagitta)));
}


bool CFitEllipse::FitDIR(CStroke Stroke, CArc &Arc,
						 bool CheckQuality, CStrokeMetrics SM)
//Returns the parameters of the ellipse arc that best fit the Stroke
//Uses the approach by Fitzgibbon et al. (DIR)
//Alternatively uses DIR/CH (DIR Convex Hull)
//Returns false if a valid fit cannnot be found

//If flag CheckQuality is true, the function checks to detect poor quality fits
//If so, tries to fit its convex hull,
//and returns false if such poor quality fits persist

//DIR is considered poor if:
//	The stroke is short and flat, and the arc is not short (does not cover a small angle)
//	The arc is greater than (overflows) the bounding box of the stroke
//	At least one of the endpoints of the stroke is too distant from the arc
//	The arc is smaller than (underflows) the boundig box
{
	//Get a DIR fit
	bool ConvexHull= false;
	CConicParams phi;

	if(!FitEllipseDIR(Stroke, phi)){
		if(FitEllipseSpeed!=1)
			return false;

		if(!GetConvexHull(Stroke))
			return false;
		ConvexHull= true;

		if(!FitEllipseDIR(Stroke, phi))
			return false;
	}

	if(!GetArc(phi, Stroke, Arc, false)){
		if(FitEllipseSpeed==0)
			return false;
		else if(ConvexHull)
			return false;
		else{
			if(!GetConvexHull(Stroke))
				return false;
			ConvexHull= true;

			if(!FitEllipseDIR(Stroke, phi) ||
			   !GetArc(phi, Stroke, Arc, false))
				return false;
		}
	}

	//Check if fit is poor
	if(CheckQuality){
		if(DisplayColours){
			Arc.ColorR= 255; Arc.ColorG= 0; Arc.ColorB= 0;
		}

		if(IsBadFitDIR(Stroke, Arc, SM)){
			if(FitEllipseSpeed==0){
				Arc.Clear();
				return false;
			}
			else if(ConvexHull)
				return false;
			else{
				if(GetConvexHull(Stroke) &&
				   FitEllipseDIR(Stroke, phi) &&
				   GetArc(phi, Stroke, Arc, false)){
					ConvexHull= true;
					if(IsBadFitDIR(Stroke, Arc, SM))
						return false;
				}
				else
					return false;
			}
		}
	}

	//Assign colour to display the arc
	if(DisplayColours){
		if(ConvexHull){
			Arc.ColorR= 0; Arc.ColorG= 155; Arc.ColorB= 255;
		}
		else{
			Arc.ColorR= 0; Arc.ColorG= 0; Arc.ColorB= 255;
		}
	}

	return true;
}


bool CFitEllipse::FastFit(CStroke Stroke, CArc &Arc, CStrokeMetrics SM)
//Calculates a fast fit automatically
//The approach seeks for an agreement between calculation time and precision

//5P is calculated first, as it is the faster
//If 5P fit is not valid, DIR approach is launched

//Valid fits may be poor quality
//Hence, the function evaluates some metrics to determine the quality of the fit
//and recalculates if needed
{
	if(FitEllipseSpeed!=2 && Fit5P(Stroke, Arc, true, true, SM))
		return true;

	else{
		Arc.Clear();

		if(ResampleStrokes){
			//Fit the resampled stroke
			CStroke RS= Stroke;
			if(RS.Resample(NumResampledPoints)){
				//if((FitEllipseSpeed!=2 ||
				//	FitEllipseSpeed==2 && !ShortStroke) &&
				//   FitDIR(RS, Arc, true, SM))
				//	return true;
				//else
				//	return false;

				return((FitEllipseSpeed!=2 ||
						FitEllipseSpeed==2 && !SM.ShortStroke) &&
					   FitDIR(RS, Arc, true, SM));
			}
			else
				return false;
		}

		else
			////Fit the original stoke
			//if(FitDIR(Stroke, Arc, true, SM)){
			//	Arc.ColorR= 0; Arc.ColorG= 225; Arc.ColorB= 255;
			//	return true;
			//}
			//else{
			//	Arc.Clear();
				return false;
			//}
	}
}


bool CFitEllipse::FitGEF(CStroke Stroke, CArc &Arc, long numStroke, long numSubStroke)
//Returns the parameters of the ellipse arc that best fit the Stroke
//Uses the approach by Szpack et al.(GEF)
//Returns false if a valid fit cannnot be found
{
	if(ResampleStrokesGEF){
		if(Stroke.GetNumPoints() > NumResampledPointsGEF)
			Stroke.Resample(NumResampledPointsGEF);
		//else
		//	return false; //Do not attempt to precisely fit strokes defined by very few points
	}

	CConicParams phi;

	if(!FitEllipseGEF(Stroke, numStroke, numSubStroke, phi))
		return false;

	if(!GetArc(phi, Stroke, Arc, false))
		return false;

	if(DisplayColours){
		if(ResampleStrokesGEF){
			Arc.ColorR= 255; Arc.ColorG= 0; Arc.ColorB= 255;
		}
		else{
			Arc.ColorR= 155; Arc.ColorG= 0; Arc.ColorB= 155;
		}
	}

	return true;
}


bool CFitEllipse::FitEllipseArc(CStroke Stroke, CArc &Arc, long numStroke, long numSubStroke)
//Returns the parameters of the ellipse arc that best fit the Stroke

//Uses four different approaches:
//	Automatic
//	Five-points (5P)
//	Fitzgibbon (DIR)
//	Szpack (GEF)

//Returns false if a valid fit cannnot be found
{
	if(!DisplayColours){
		Arc.ColorR= 0; Arc.ColorG= 0; Arc.ColorB= 0;
	}

	switch (FitEllipseApproach){
		case 0:{  //Automatic
			CStrokeMetrics SM;
			GetStrokeMetrics(Stroke, SM);

			//Automatic fit sequence
			if(FastFit(Stroke, Arc, SM))
				return true;
			else if(FitEllipseSpeed==2 && FitGEF(Stroke, Arc, numStroke, numSubStroke))
				return true;
			else{
				Arc.Clear();
				return false;
			}
		}
		break;

		case 1:{  //5P, Five-points
			CStrokeMetrics SM;
			if(CheckQuality5P)
				GetStrokeMetrics(Stroke, SM);
			if(!Fit5P(Stroke, Arc, UseConvexHull5P, CheckQuality5P, SM))
				return false;
		}
		break;

		case 2:{  //DIR, Fitzgibbon
			if(ResampleStrokes)
				Stroke.Resample(NumResampledPoints);  //Resample to reduce calculation time

			CStrokeMetrics SM;
			if(CheckQualityDIR)
				GetStrokeMetrics(Stroke, SM);
			if(!FitDIR(Stroke, Arc, CheckQualityDIR, SM))
				return false;
		}
		break;

		case 3:{  //GEF, Szpack
			if(!FitGEF(Stroke, Arc, numStroke, numSubStroke))
				return false;
		}
		break;
	}//end switch

	return true;
}


double CFitEllipse::GetStrokeArcWidth(CStroke &Stroke, CArc Arc)
//Returns the width of the bounding strip containing the stroke
//Returns the outside band width (t_Out) and the inside band width (t_In)

//The stroke is compared against the elliptic arc that better fits it

//At this end, it is assumed that an ellipse
//is a circle that has been stretched in one direction
//Hence, the stroke is rotated and scaled
//using the same factors that should convert the fitted elliptic arc
//into a unit radius circle

//Then the points of the transformed stroke more distant from the unit radius
//are considered to be the most distant from the ellipse
{
	CStroke S= Stroke;

	//Transform the stroke into "circle system"
	POINT2D centre (Arc.Centre_x, Arc.Centre_y);
	S.Rotate(centre, -Arc.Rotation);
	POINT2D scale (1/Arc.Radius_a, 1/Arc.Radius_b);
	S.Scale(centre, scale);

	//Find the point of the stroke more distant from the ellipse
	double maxdistOut= 0., maxdistIn= 0.;
	long numworstOut= 0, numworstIn= 0;

	for (long i=0; i<(long)Stroke.GetNumPoints(); i++){
		POINT2D pp= S.GetPoint(i).GetPosition();
		pp.x-= Arc.Centre_x;
		pp.y-= Arc.Centre_y;
		double currentdist= sqrt(pp.x*pp.x + pp.y*pp.y) -1;

		if(currentdist > maxdistOut){
			maxdistOut= currentdist;
			numworstOut= i;
		}
		if(currentdist < maxdistIn){
			maxdistIn= currentdist;
			numworstIn= i;
		}
	}

	//Get the distance from the worst outside point of the stroke to the ellipse
	POINT2D pw= Stroke.GetPoint(numworstOut).GetPosition();
	Stroke.worstOut_x= pw.x;   //Used for display
	Stroke.worstOut_y= pw.y;   //Used for display

	//Transform to "ellipse system"
	pw= pw - centre;
	pw= pw.Rotate(-Arc.Rotation);

	POINT2D pe= ClosestPointInEllipse(Arc.Radius_a, Arc.Radius_b, pw);
	Stroke.t_Out= MODULE_2D(pw, pe);

	//Get the distance from the worst inside point of the stroke to the ellipse
	pw= Stroke.GetPoint(numworstIn).GetPosition();
	Stroke.worstIn_x= pw.x;     //Used for display
	Stroke.worstIn_y= pw.y;     //Used for display

	//Transform to "ellipse system"
	pw= pw - centre;
	pw= pw.Rotate(-Arc.Rotation);

	pe= ClosestPointInEllipse(Arc.Radius_a, Arc.Radius_b, pw);
	Stroke.t_In= MODULE_2D(pw, pe);

	return (Stroke.t_Out + Stroke.t_In);
}


long CFitEllipse::SmoothStroke(CStroke &Stroke,
							   CArc Arc,
							   double ToleranceWidth,
							   double widthInitial)
//Returns a "smoothed" stroke
//Smoothing consists in removing odd point of the stroke

//Also returns the number of smoothing stages

//The goal is removing high frequency oscillations
//As they are small, repeatedly removing odd stroke points makes them eventually dissapear

//Smoothing ends when the smoothed stroke:
//	1. does contain no more corners, or
//	2. becomes significantly marrow than the original one

//Odd point that are too distant from ther neighbours are not removed
//First and last points are not removed
{
	long NumSmoothingSteps= 0;

	if(Stroke.NumCorners() <= 0)
		return NumSmoothingSteps;

	while(true){
		CStroke TempStroke= Stroke;
		if(!TempStroke.Filter(2*ToleranceWidth))    // Tuning Parameter
			break;  //No more filtering allowed
		NumSmoothingSteps++;

		double widthSmoothed= GetStrokeArcWidth(TempStroke, Arc);

		if(abs(widthInitial-widthSmoothed) >= ToleranceWidth)    // Tuning Parameter
			break;  // Already significantly marrow than the original one
		else{
			Stroke= TempStroke;
			if(Stroke.FindCorners(StrokeCornersApproach) <= 0)
				break;  //No more corners to smooth
		}
	}

	return NumSmoothingSteps;
}


double CFitEllipse::GetMeritEllipse(CStroke &Stroke, long numStroke, long numSubStroke, CArc &Arc)
//Returns the merit of the stroke as an elliptic arc

//Display information is also returned in the Stroke (worstOut and worstIn)

//Sparse strokes (those with less than five points) are trivially labelled as non-elliptic
//as they represent polygonal lines, not arcs

//The merit equals 1 if the shape is an ellipse,
//and equals 0 when the shape is highly non-elliptical

//Stroke is accepted as representing an elliptic arc if:
//		1. Fits into a minimum tolerance band,
//or, alternatively:
//		2. Its smoothed shape fits into a maximum (or "relaxed") tolerance band.

//3.Stroke is smoothed to remove oscillations,
//But the merit of the smoothed stroke is reduced by ArcSmoothingPenalty

//Also returns the parameters of the elliptic arc:
//	the centre
//	the semi-axes lengths
//	the counterclockwise angle of rotation from the x-axis to the major axis
//	the foci
{
	//Reject sparse strokes as candidate to depict elliptic arcs
	//----------------------------------------------------------
	if(Stroke.GetNumPoints() < 5)
		return 0.0;

	//Fit stroke into a temptative ellipse
	//------------------------------------
	if(!FitEllipseArc(Stroke, Arc, numStroke, numSubStroke))
		return 0.0;

	//Check the minimum tolerance criterion
	//-------------------------------------
	double MeritArc= 1.0;
	double Tol= GetStrokeArcWidth(Stroke, Arc);

	double TolMin= ArcTolMin * Stroke.GetLength(10);

	if(Tol < TolMin)
		return MeritArc;

	//Smooth oscillations
	//-------------------
	CStroke SmoothedStroke= Stroke;
	double TolSmooth= Stroke.GetLength(10)/100;    // Tuning Parameters
	long NumSmoothingSteps= SmoothStroke(SmoothedStroke, Arc, TolSmooth, Tol);

	//Check the maximum (relaxed) tolerance criterion
	//-----------------------------------------------
	Tol= GetStrokeArcWidth(SmoothedStroke, Arc);

	double TolMax= ArcTolMax * SmoothedStroke.GetLength(10);

	//Assign merit
	if(Tol < TolMin)
		MeritArc= 1.0;
	else if(Tol > TolMax)
		MeritArc= 0.0;
	else
		MeritArc= 1.0 - (Tol-TolMin)/(TolMax-TolMin);

	//Reduce merit for oscillating arcs
	//---------------------------------
	MeritArc= std::max(0.0, (MeritArc - ArcSmoothingPenalty*NumSmoothingSteps));     // Tuning Parameter

	return MeritArc;
}
