//+-----------------------------------------------------------------------------+
//| Copyright (C) 2013 REGEO (http://www.regeo.uji.es/)
//|
//| This program is free software
//| You can redistribute it and/or modify it under the terms of the GNU General Public License,
//| as published by the Free Software Foundation.
//| 
//| This program is distributed in the hope that it will be useful,
//| but WITHOUT ANY WARRANTY; without even the implied warranty of
//| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//+----------------------------------------------------------------------------*/
#include <cassert>     // assert
#include <cmath>    // sqrt
#include <algorithm>
#include <vector>
#include "Tools_Geometry.h"

#include "Stroke.h"
#include "StrokeCornerFinder.h"

// Database for Sketches, strokes and points
// Includes the definition of the function that translates from this database to REFER databases

//----------------------------------------------------------------------
// POINTS IN STROKE
// Stores points that belong to a stroke
// Storage includes location (a pair or real coordinates)
// plus sequence (a time parameter)

CStrokePoint::CStrokePoint()
{
}


CStrokePoint::CStrokePoint(double cx, double cy, double ctime)
// Stores coordinates of one single point sampled when moving the cursor in sketch mode 
:
x (cx),
y (cy),
time (ctime)
{
}


CStrokePoint::CStrokePoint(const CStrokePoint &IP)
// Copy constructor
:
x (IP.x),
y (IP.y),
time (IP.time)
{
}

CStrokePoint::~CStrokePoint()
{
}


void CStrokePoint::PutPosition(double cx, double cy)
// Updates coordinates of one single point sampled when moving the cursor in sketch mode 
{
	x= cx;
	y= cy;
}


POINT2D CStrokePoint::GetPosition()
// Returns coordinates of one single point sampled when moving the cursor in sketch mode 
{
	POINT2D c;
	c.x= x;
	c.y= y;

	return c;
}


void CStrokePoint::PutTime(double ctime)
// Updates sampling time of one single point sampled when moving the cursor in sketch mode 
{
	time= ctime;
}


double CStrokePoint::GetTime()
// Returns sampling time of one single point sampled when moving the cursor in sketch mode 
{
	return time;
}


//----------------------------------------------------------------------
// STROKES


CStroke::CStroke()
// Stores information of a single stroke:
//    StrokePoints  is a vector that contains pointers to the sequence of points
//                  that belong to the same sketched stroke

// Parameters linked to its vectorization are saved inside the stroke
{
	positions.clear();
	grouped= false;
	segmented= false;

	MeritLine= 0.0;
	MeritArc= 0.0;
	NumLine= -1;
	NumArc= -1;

	worstOut_x= 0.;
	worstOut_y= 0.;
	t_Out= 0.;
	worstIn_x= 0.;
	worstIn_y= 0.;
	t_In= 0.;
}


CStroke::CStroke(const CStroke &Stroke)
// Copy constructor
:
StrokePoints (Stroke.StrokePoints),

positions (Stroke.positions),
grouped (Stroke.grouped),
segmented (Stroke.segmented),

MeritLine (Stroke.MeritLine),
NumLine (Stroke.NumLine),

MeritArc (Stroke.MeritArc),
NumArc (Stroke.NumArc),

worstOut_x (Stroke.worstOut_x),   //Used for display
worstOut_y (Stroke.worstOut_y),   //Used for display
t_Out (Stroke.t_Out),
worstIn_x (Stroke.worstIn_x),   //Used for display
worstIn_y (Stroke.worstIn_y),   //Used for display
t_In (Stroke.t_In)
{
}


CStroke::~CStroke()
{
}


CStroke CStroke::operator+(const CStroke& right) const
//Overload of + operator for CStroke
{
	CStroke Sum;
	Sum.StrokePoints= this->StrokePoints;
	for(long i= 0; i<(long)right.StrokePoints.size(); i++){
		Sum.StrokePoints.push_back(right.StrokePoints[i]);
	}
	return Sum;
}


void CStroke::Clear()
// Clears the stroke
{
	StrokePoints.clear();
	positions.clear();
	grouped= false;
	segmented= false;

	ClearMerits();
}


void CStroke::AddPoint(CStrokePoint Point)
// Adds a new point to the end of the stroke
{
	StrokePoints.push_back(Point);
}


CStrokePoint CStroke::GetPoint(long iPoint)
// Returns the point located in position iPoint
{
	assert(iPoint>=0 && iPoint< (long)StrokePoints.size());
	return StrokePoints[iPoint];
}


void CStroke::PutPoint(long iPoint, CStrokePoint Point)
// Updates the point in position iPoint
{
	assert(iPoint>=0 && iPoint<(long)StrokePoints.size());
	StrokePoints[iPoint]= Point;
}


void CStroke::RemovePoint(long iPoint)
// Removes the point located in position iPoint
{
	assert(iPoint>=0 && iPoint< (long)StrokePoints.size());
	StrokePoints.erase(StrokePoints.begin()+iPoint);
}


CStroke CStroke::GetSubStroke(long PositionI, long PositionF)
//Returns a substroke starting at PositionI and ending at PositionF
{
	assert(PositionF >= PositionI);
	assert(PositionI >= 0);
	assert(PositionF <= (long)StrokePoints.size());

	CStroke SubStroke;

	for (long i=PositionI; i<=PositionF; i++){
		SubStroke.AddPoint(GetPoint(i));
	}

	return SubStroke;
}


long CStroke::GetNumPoints()
// Returns the number of point in the stroke
{
	return StrokePoints.size();
}


RECTANGLE2D CStroke::GetBounds()
// Finds the bounding rectangle of the current stroke
{
	RECTANGLE2D bound;

	long n= StrokePoints.size();
	if (n==0){
		bound.left= bound.right= bound.bottom= bound.top= 0;
		return bound;
	}

	POINT2D c= StrokePoints[0].GetPosition();
	bound.left  = bound.right = c.x;
	bound.bottom= bound.top   = c.y;

	for (long ipoint=1; ipoint<n; ipoint++){
		c= StrokePoints[ipoint].GetPosition();
		bound.left  = std::min(bound.left, c.x);
		bound.right = std::max(bound.right, c.x);
		bound.bottom= std::min(bound.bottom, c.y);
		bound.top   = std::max(bound.top, c.y);
	}

	return bound;
}


double CStroke::GetLength()
//Returns the length of the stroke
//calculated adding the lengths of the chords
//that connect all the consecutive points
{
	double Length= 0.0;
	POINT2D J= StrokePoints[0].GetPosition();
	long numP= StrokePoints.size();
	for (long j=1; j<numP; j++){
		POINT2D JJ= StrokePoints[j].GetPosition();
		Length+= MODULE_2D(J, JJ);
		J= JJ;
	}

	return Length;
}


double CStroke::GetLength(long numIntervals)
//Returns the approximate length of the stroke without oscillations

//The length is calculated considering only the lenghts
//of "numIntervals" chords

//A small value of numIntervals (typically 10) is used
//to find the approximate length of the undulating arc
//that should result after removing high frequency oscillations
//from the original arc

//A too small value of numIntervals would also remove undulations
//and even the underlying curvature of the stroke
{
	double Length= 0.0;
	POINT2D J= StrokePoints[0].GetPosition();
	long numP= StrokePoints.size();
	long delta= std::max((long) 1, (long) numP/numIntervals);
	long j;
	for (j=1; j<numP; j=j+delta){
		POINT2D JJ= StrokePoints[j].GetPosition();
		Length+= MODULE_2D(J, JJ);
		J= JJ;
	}
	if(j<numP-1){
		POINT2D JJ= StrokePoints[numP-1].GetPosition();
		Length+= MODULE_2D(J, JJ);
	}

	return Length;
}


double CStroke::GetSlope180()
//Returns the slope of the segment that connects the endpoints of the stroke

//Slope is returned in the range [0, PI)

//Initial point is the tail
//Final point is the head

//Returns null slope if both endpoints are coincident
{
	POINT2D ini= GetPoint(0).GetPosition();
	POINT2D end= GetPoint(GetNumPoints()-1).GetPosition();
	if(MODULE_2D(ini, end) == 0.)
		return 0.;

	//double slope90= atan2(end.y-ini.y, end.x-ini.x);

	//if(ini.x == end.x)
	//	if(ini.y < end.y)
	//		return slope90;
	//	else
	//		return (slope90 + NUMBER_PI);
	//else if(ini.x < end.x)
	//	return slope90;
	//else
	//	return (slope90 + NUMBER_PI);


	double slope180= atan2(end.y-ini.y, end.x-ini.x);
	if(slope180 < 0) slope180+= NUMBER_PI;
	return slope180;
}


double CStroke::GetMaxSegment()
//Returns the length of the longer segment
//connecting two consecutive points of the stroke
{
	double maxLength= 0.0;
	POINT2D J= StrokePoints[0].GetPosition();
	long numP= StrokePoints.size();
	for (long j=1; j<numP; j++){
		POINT2D JJ= StrokePoints[j].GetPosition();
		maxLength= std::max(maxLength, MODULE_2D(J, JJ));
		J= JJ;
	}

	return maxLength;
}


POINT2D CStroke::GetCentroid()
//Returns the centroid of the stroke
{
	double sumx= 0, sumy= 0, sumx2= 0, sumy2= 0, sumxy= 0;
	long numP= StrokePoints.size();
	for (long j=0; j<numP; j++){
		POINT2D J= StrokePoints[j].GetPosition();
		sumx+= J.x;
		sumy+= J.y;
	}
	POINT2D Centroid(sumx/numP, sumy/numP);
	return Centroid;
}


std::vector<std::vector<double>> CStroke::GetNormaliseMatrix()
// Returns the matrix to normalise the points of a stroke
// so that their mean distance from their centroid is sqrt(2)
// and the centroid is the origin

// This process typically improves the conditioning of any equations
//  used to solve homographies, fundamental matrices etc.
{
	//Centroid
	POINT2D c= this->GetCentroid();

	//Scale so that the mean distance from the centroid to a point equals sqrt(2)
	double meandist= 0.;
	long numP= StrokePoints.size();
	for(long i=0; i<numP; i++){
		POINT2D I= StrokePoints[i].GetPosition();
		double px= I.x - c.x;
		double py= I.y - c.y;
		meandist+= sqrt(px*px + py*py);
	}
	meandist/= (double)numP;
	double scale = sqrt(2.)/meandist;

	//Normalisation matrix
	std::vector<std::vector<double>> N;
	N[0][0] = scale;   N[0][1] = 0.;      N[0][2] = -scale*c.x;
	N[1][0] = 0.;      N[1][1] = scale;   N[1][2] = -scale*c.y;
	N[2][0] = 0.;      N[2][1] = 0.;      N[2][2] = 1.;

	return N;
}


bool CStroke::IsClockwise()
//Returns true if the points of the stroke are sequenced clockwise
//and false if they are sequenced counter-clockwise

//(Calculation is done by the shoelace area formula)
{
	//Detect if stroke points are in clockwise order
	long numP= (long)StrokePoints.size();
	double signedArea= 0;
	for (long i=0; i<numP; i++){
		POINT2D Current= StrokePoints[i].GetPosition();
		POINT2D Next= StrokePoints[(i+1)%numP].GetPosition();
		signedArea+= (Next.x-Current.x) * (Next.y+Current.y);
	}

	return(signedArea > 0);
}



void CStroke::GetConvexHull()
//Modifies the input stroke to return its convex hull

//First determine if the stroke is clockwise or counter-clockwise
//(by the shoelace area formula)

//Then concave points are sequentially detected and removed
{
	//Detect if stroke points are in clockwise order
	long numP= (long)StrokePoints.size();
	double signedArea= 0;
	for (long i=0; i<numP; i++){
		POINT2D Current= StrokePoints[i].GetPosition();
		POINT2D Next= StrokePoints[(i+1)%numP].GetPosition();
		signedArea+= (Next.x-Current.x) * (Next.y+Current.y);
	}

	bool Clockwise= IsClockwise();

	//Sequentially remove concave points
	//(points to the right of the line defined by
	//the previous and next points if sequence is clockwise,
	//or to the left is sequence is counter clockwise)
	bool repeat= true;
	POINT2D Previous= GetPoint(0).GetPosition();
	while(repeat){
		repeat= false;
		for (long i=1; i<(long)StrokePoints.size()-1; i++){
			POINT2D Current= StrokePoints[i].GetPosition();
			POINT2D Next= StrokePoints[i+1].GetPosition();

			double product= (Next.x-Previous.x)*(Current.y-Previous.y) - (Next.y-Previous.y)*(Current.x-Previous.x);

			if((Clockwise && product < 0.) || (!Clockwise && product >0.)){
				RemovePoint(i);
				i--;
				repeat= true;
			}
			else
				Previous= Current;
		}
	}
}


void CStroke::Translate(POINT2D distance)
//Translates the stroke
{
	for (long i=0; i<(long)StrokePoints.size(); i++){
		POINT2D P= StrokePoints[i].GetPosition();
		//double time= StrokePoints[i].GetTime();

		P.x += distance.x;
		P.y += distance.y;

		StrokePoints[i].PutPosition(P.x, P.y);
		//StrokePoints[i].PutTime(time);
	}
}


void CStroke::Scale(POINT2D Centre, POINT2D Scale)
//Scales the stroke around Centre
{
	for (long i=0; i<(long)StrokePoints.size(); i++){
		POINT2D P= StrokePoints[i].GetPosition();
		//double time= StrokePoints[i].GetTime();

		POINT2D Pr;
		Pr.x= Centre.x + (P.x-Centre.x)*Scale.x;
		Pr.y= Centre.y + (P.y-Centre.y)*Scale.y;

		StrokePoints[i].PutPosition(Pr.x, Pr.y);
		//StrokePoints[i].PutTime(time);
	}
}


void CStroke::Rotate(POINT2D Centre, double angle)
//Rotates the stroke and angle alpha around Centre
{
	for (long i=0; i<(long)StrokePoints.size(); i++){
		POINT2D P= StrokePoints[i].GetPosition();
		//double time= StrokePoints[i].GetTime();

		POINT2D Pr;
		Pr.x= Centre.x+((P.x-Centre.x)*cos(angle)-(P.y-Centre.y)*sin(angle));
		Pr.y= Centre.y+((P.x-Centre.x)*sin(angle)+(P.y-Centre.y)*cos(angle));

		StrokePoints[i].PutPosition(Pr.x, Pr.y);
		//StrokePoints[i].PutTime(time);
	}
}


void CStroke::CleanEnds()
//Removes points overlapped with the first and last points of the stroke

//Those points result from the user moving the pointer without displacing it
{
	bool remove= true;
	while(remove){
		remove= false;
		POINT2D First= StrokePoints[0].GetPosition();
		POINT2D Next= StrokePoints[1].GetPosition();
		if(MODULE_2D(First, Next) == 0){
			RemovePoint(0);
			remove= true;
		}
	}

	long n= StrokePoints.size()-1;
	remove= true;
	while(remove){
		remove= false;
		POINT2D Previous= StrokePoints[n-1].GetPosition();
		POINT2D Last= StrokePoints[n].GetPosition();
		if(MODULE_2D(Previous, Last) == 0){
			RemovePoint(n);
			n--;
			remove= true;
		}
	}
}


void CStroke::Filter(long FilterGap)
//Removes one of every "FilterGap" points of the stroke
{
	long SizeI= (long)StrokePoints.size();
	for (long i=SizeI-2; i>0; i-=FilterGap){
		RemovePoint(i);
	}
}


bool CStroke::Filter(double maxDistance)
//Removes alternate points of the stroke
//But only if the distance between consecutive remaining points
//is smaller than "maxDistance"
{
	bool filtered= false;
	long Last= (long)StrokePoints.size()-1;

	for (long i=Last-1; i>=2; i--){
		POINT2D Previous= StrokePoints[i+1].GetPosition();
		POINT2D Current= StrokePoints[i].GetPosition();
		if(MODULE_2D(Previous, Current) < maxDistance){
			RemovePoint(i);
			filtered= true;
			i--;
		}
	}
	return filtered;
}


bool CStroke::Resample(long NumResampledPoints)
//Resamples the stroke to get a new stroke with
//NumPoints approximately equally distanced

//WARNING: note that chords have dissimilar lengths
//(even if archs are the same length),
//when curvature changes abruptly
{
	if((long)StrokePoints.size() <= NumResampledPoints ||
	   NumResampledPoints < 3)
		return false;

	//Add first point
	std::vector<CStrokePoint> New;
	New.push_back(StrokePoints[0]);

	//Loop to add intermediate points
	double Length= GetLength();
	double Interval= Length/(NumResampledPoints-1);

	long Last= (long)StrokePoints.size()-1;
	double accLength= 0.;
	POINT2D Previous= StrokePoints[0].GetPosition();

	for (long i=1; i<=Last; i++){
		POINT2D Current= StrokePoints[i].GetPosition();
		accLength+= MODULE_2D(Previous, Current);
		if(accLength > Interval){
			//Interpolate current point
			double delta= (accLength-Interval)/MODULE_2D(Previous, Current);
			Current.x-= (Current.x - Previous.x)*delta;
			Current.y-= (Current.y - Previous.y)*delta;
			double Current_time= StrokePoints[i].GetTime();
			Current_time-= (Current_time-StrokePoints[i-1].GetTime())*delta;
			CStrokePoint sp(Current.x, Current.y, Current_time);
			New.push_back(sp);
			accLength= 0.;
			i--;
		}
		else if(accLength == Interval){
			New.push_back(StrokePoints[i]);
			accLength= 0.;
		}
		Previous= Current;
	}

	//Add last point
	if(abs(Interval-accLength) < Interval/100.)
		New.push_back(StrokePoints[Last]);

	StrokePoints= New;
	assert(StrokePoints.size() == NumResampledPoints);
	return true;
}


long CStroke::FindCorners(long StrokeCornersApproach)
//Finds the corners of the stroke
//and returns the number of corners found

//The two endponts are not counted as corners

//Corners are stored in vector "positions"
{
	positions.clear();

	CStrokeCornerFinder CF;
	CF.FindStrokeCorners(*this, positions, StrokeCornersApproach);

	return (positions.size()-2);
}


long CStroke::NumCorners()
//Returns the number of corners of the stroke
{
	long n= positions.size()-2;
	//if (n<0)
		//AfxMessageBox("Corners not calculated");

	return n;
}


void CStroke::ClearMerits()
// Clears the stroke merits
{
	MeritLine= 0.0;
	MeritArc= 0.0;
	NumLine= -1;
	NumArc= -1;

	worstOut_x= 0.;
	worstOut_y= 0.;
	t_Out= 0.;
	worstIn_x= 0.;
	worstIn_y= 0.;
	t_In= 0.;
}
