//+-----------------------------------------------------------------------------+
//| Copyright (C) 2013 REGEO (http://www.regeo.uji.es/)
//|
//| This program is free software
//| You can redistribute it and/or modify it under the terms of the GNU General Public License,
//| as published by the Free Software Foundation.
//| 
//| This program is distributed in the hope that it will be useful,
//| but WITHOUT ANY WARRANTY; without even the implied warranty of
//| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//+----------------------------------------------------------------------------*/

#include <cmath>
#include <cassert>
#include <vector>

#include "Tools_Geometry.h"


POINT2D::POINT2D()
:x (0),
 y (0)
{
}


POINT2D::POINT2D(const POINT2D& DP)
:x (DP.x),
 y (DP.y)
{
}


POINT2D::POINT2D(double cx, double cy)
:x (cx),
 y (cy)
{
}


POINT2D::~POINT2D()
{
}


bool POINT2D::operator==(const POINT2D& right) const
//Overload of == operator for POINT2D
{
	return (this->x==right.x && this->y==right.y);
}


bool POINT2D::operator!=(const POINT2D& right) const
//Overload of != operator for POINT2D
{
	return (this->x!=right.x || this->y!=right.y);
}


POINT2D& POINT2D::operator+=(const POINT2D& right)
//Overload of += operator for POINT2D
{
	this->x+= right.x;
	this->y+= right.y;
	return *this;
}


POINT2D POINT2D::operator+(const POINT2D& right) const
//Overload of + operator for POINT2D
{
	POINT2D p(this->x+right.x, this->y+right.y);
	return p;
	//return POINT2D (*this) += right;
}


POINT2D& POINT2D::operator-=(const POINT2D& right)
//Overload of -= operator for POINT2D
{
	this->x-= right.x;
	this->y-= right.y;
	return *this;
}


POINT2D POINT2D::operator-(const POINT2D& right) const
//Overload of - operator for POINT2D
{
	POINT2D p(this->x-right.x, this->y-right.y);
	return p;
	//return POINT2D (*this) -= right;
}


POINT2D& POINT2D::operator*=(const POINT2D& right)
//Overload of *= operator for POINT2D
{
	this->x*= right.x;
	this->y*= right.y;
	return *this;
}


POINT2D POINT2D::operator*(const POINT2D& right) const
//Overload of * operator for POINT2D
{
	POINT2D p(this->x*right.x, this->y*right.y);
	return p;
	//return POINT2D (*this) *= right;
}


POINT2D& POINT2D::operator/=(const POINT2D& right)
//Overload of /= operator for POINT2D
{
	assert(right.x!=0);
	assert(right.y!=0);
	this->x/= right.x;
	this->y/= right.y;
	return *this;
}


POINT2D POINT2D::operator/(const POINT2D& right) const
//Overload of / operator for POINT2D
{
	assert(right.x!=0);
	assert(right.y!=0);
	POINT2D p(this->x/right.x, this->y/right.y);
	return p;
	//return POINT2D (*this) /= right;
}


POINT2D POINT2D::Rotate(POINT2D Center, double beta)
//Returns coordinates of the point rotated and angle "beta" around the point "center"
{
	POINT2D PRot;
	PRot.x= Center.x+((x-Center.x)*cos(beta)-(y-Center.y)*sin(beta));
	PRot.y= Center.y+((x-Center.x)*sin(beta)+(y-Center.y)*cos(beta));
	return PRot;
}


POINT2D POINT2D::Rotate(double beta)
//Returns coordinates of the point rotated and angle "beta" around the origin
{
	POINT2D PRot;
	PRot.x= x*cos(beta)-y*sin(beta);
	PRot.y= x*sin(beta)+y*cos(beta);
	return PRot;
}


SEGMENT2D::SEGMENT2D()
:p0 (0,0),
 p1 (0,0)
{
}


SEGMENT2D::SEGMENT2D(const SEGMENT2D& DS)
:p0 (DS.p0),
 p1 (DS.p1)
{
}


SEGMENT2D::SEGMENT2D(POINT2D cp0, POINT2D cp1)
:p0 (cp0),
 p1 (cp1)
{
}


SEGMENT2D::~SEGMENT2D()
{
}


void SEGMENT2D::Rotate(POINT2D Center, double beta)
//Returns coordinates of the point rotated and angle "beta" around the point "center"
{
	p0= p0.Rotate(Center, beta);
	p1= p1.Rotate(Center, beta);
}


bool SEGMENT2D::operator==(const SEGMENT2D& right) const
//Overload of == operator for SEGMENT2D
{
	return (this->p0==right.p0 && this->p1==right.p1);
}


bool SEGMENT2D::operator!=(const SEGMENT2D& right) const
//Overload of != operator for SEGMENT2D
{
	return (this->p0!=right.p0 || this->p1!=right.p1);
}


RECTANGLE2D::RECTANGLE2D()
:left (0),
 right (0),
 bottom (0),
 top (0)
{
}


RECTANGLE2D::RECTANGLE2D(const RECTANGLE2D& DR)
:left (DR.left),
 right (DR.right),
 bottom (DR.bottom),
 top (DR.top)
{
}


RECTANGLE2D::RECTANGLE2D(double cleft, double cright, double cbottom, double ctop)
:left (cleft),
 right (cright),
 bottom (cbottom),
 top (ctop)
{
}


RECTANGLE2D::~RECTANGLE2D()
{
}


double GetDistancePointToSegment(POINT2D p0, POINT2D p1, POINT2D V,
								 POINT2D &Intersection)
//Gets the distance from an external point "V" to a segment defined by p0 and p1
//Gets also the intersection of p0-p1 with the minimum distance segment

//Returns the distance and the intersection point
{
	if(MODULE_2D(p0, p1) == 0)
		return MODULE_2D(p0, V);

	//Get the intersection point
	double u= (V.x - p0.x)*(p1.x - p0.x) + (V.y - p0.y)*(p1.y - p0.y);
	u/= ((p1.x - p0.x)*(p1.x - p0.x) + (p1.y - p0.y)*(p1.y - p0.y));
	Intersection.x = p0.x + ((p1.x - p0.x)*u);
	Intersection.y = p0.y + ((p1.y - p0.y)*u);

	//Get the distance from the external point to the intersection point
	return MODULE_2D(Intersection, V);
}



double GetAngle_2PI(POINT2D T, POINT2D H)
//Gets the angle (in radians) between
//	the segment defined by T and H
//	and the X axis

//The edge is oriented: T is the tail and H is the head,
//Returns a value in the range [0, 2PI)
{
	double length= MODULE2D(T.x, T.y, H.x, H.y);
	assert(length > 0);

	double Angle;
	if (H.y >= T.y){   // [0, 180]
		Angle= acos((H.x-T.x)/length);
	}
	else{   // (180, 360]
		Angle= acos((T.x-H.x)/length);
		Angle+= NUMBER_PI;
	}
	Angle= std::max(Angle, 0.);
	Angle= std::min(Angle, 2.*NUMBER_PI);
	if (fabs(Angle-2*NUMBER_PI)<Accuracy)
		Angle= 0.0;

	return Angle;
}


double GetAngle_2PI(SEGMENT2D S)
//Gets the angle (in radians) between the segment S and the X axis

//The edge is oriented: S.p0 is the tail and S.p1 is the head,
//Returns a value in the range [0, 2PI)
{
	return GetAngle_2PI(S.p0, S.p1);
}


double GetAngle3Points(POINT2D P1, POINT2D P2, POINT2D P3)
//Gets the angle (in radians) between the two segments defined by P1-P2 and P2-P3

//Returns a value in the range [0, PI]
{
	double Angle= fabs(GetAngle_2PI(P2, P1) - GetAngle_2PI(P2, P3));
	if (Angle>NUMBER_PI)
		Angle= 2*NUMBER_PI-Angle;

	return Angle;
}


double GetDistancePointToSegment(POINT2D p0, POINT2D p1, POINT2D V)
//Gets the distance from an external point "V" to a segment defined by p0 and p1
{
	if(MODULE_2D(p0, p1) == 0)
		return MODULE_2D(p0, V);

	return fabs(((V.x - p0.x)*(p1.y - p0.y) - (V.y - p0.y)*(p1.x - p0.x)) / MODULE_2D(p0, p1));
}



RECTANGLE2D GetBounds(std::vector <POINT2D> Points)
// Finds the bounding rectangle of the set of Points
{
	RECTANGLE2D bound;
	bound.left  = Points[0].x;
	bound.right = Points[0].x;
	bound.bottom= Points[0].y;
	bound.top   = Points[0].y;

	for (long ipoint=1; ipoint<(long)Points.size(); ipoint++){
		bound.left  = std::min(bound.left, Points[ipoint].x);
		bound.right = std::max(bound.right, Points[ipoint].x);
		bound.bottom=  std::min(bound.bottom, Points[ipoint].y);
		bound.top   =  std::max(bound.top, Points[ipoint].y);
	}

	return bound;
}

